import logging
import os
from abc import abstractmethod
from typing import Dict, Optional, Union

import torch
import torchvision.models as tvm

from ludwig.api_annotations import DeveloperAPI
from ludwig.constants import IMAGE
from ludwig.encoders.image.base import ImageEncoder
from ludwig.encoders.registry import register_encoder
from ludwig.schema.encoders.image.torchvision import (
    TVAlexNetEncoderConfig,
    TVConvNeXtEncoderConfig,
    TVDenseNetEncoderConfig,
    TVEfficientNetEncoderConfig,
    TVGoogLeNetEncoderConfig,
    TVInceptionV3EncoderConfig,
    TVMaxVitEncoderConfig,
    TVMNASNetEncoderConfig,
    TVMobileNetV2EncoderConfig,
    TVMobileNetV3EncoderConfig,
    TVRegNetEncoderConfig,
    TVResNetEncoderConfig,
    TVResNeXtEncoderConfig,
    TVShuffleNetV2EncoderConfig,
    TVSqueezeNetEncoderConfig,
    TVSwinTransformerEncoderConfig,
    TVVGGEncoderConfig,
    TVViTEncoderConfig,
    TVWideResNetEncoderConfig,
)
from ludwig.utils.image_utils import register_torchvision_model_variants, torchvision_model_registry, TVModelVariant

logger = logging.getLogger(__name__)


@DeveloperAPI
class TVBaseEncoder(ImageEncoder):
    def __init__(
        self,
        model_variant: Union[str, int] = None,
        use_pretrained: bool = True,
        saved_weights_in_checkpoint: bool = False,
        model_cache_dir: Optional[str] = None,
        trainable: bool = True,
        **kwargs,
    ):
        super().__init__()

        logger.debug(f" {self.name}")
        # map parameter input feature config names to internal names
        self.model_variant = model_variant
        self.use_pretrained = use_pretrained
        self.model_cache_dir = model_cache_dir

        # remove any Ludwig specific keyword parameters
        kwargs.pop("encoder_config", None)
        kwargs.pop("type", None)
        kwargs.pop("skip", None)

        # cache pre-trained models if requested
        # based on https://github.com/pytorch/vision/issues/616#issuecomment-428637564
        if self.model_cache_dir is not None:
            os.environ["TORCH_HOME"] = self.model_cache_dir

        # retrieve function to create requested model
        self.create_model = torchvision_model_registry[self.torchvision_model_type][
            self.model_variant
        ].create_model_function

        # get weight specification
        if use_pretrained and not saved_weights_in_checkpoint:
            weights_specification = torchvision_model_registry[self.torchvision_model_type][
                self.model_variant
            ].model_weights.DEFAULT
            logger.info(
                f"Instantiating torchvision image encoder '{self.torchvision_model_type}' with pretrained weights: "
                f"{weights_specification}."
            )
        else:
            weights_specification = None
            if saved_weights_in_checkpoint:
                logger.info(
                    f"Instantiating torchvision image encoder: '{self.torchvision_model_type}' "
                    "with weights saved in the checkpoint."
                )
            else:
                logger.info(
                    f"Instantiating torchvision image encoder: '{self.torchvision_model_type}' "
                    "with no pretrained weights."
                )

        # get torchvision transforms object
        transforms_obj = torchvision_model_registry[self.torchvision_model_type][
            self.model_variant
        ].model_weights.DEFAULT.transforms()

        # capture key attributes from torchvision transform for later use
        self.num_channels = len(transforms_obj.mean)
        self.normalize_mean = transforms_obj.mean
        self.normalize_std = transforms_obj.std
        self.crop_size = transforms_obj.crop_size

        logger.debug(f"  {self.torchvision_model_type}")
        # create pretrained model with pretrained weights or None for untrained model
        self.model = self.create_model(weights=weights_specification, **kwargs)

        # remove final classification layer
        self._remove_softmax_layer()

        # freeze parameters if requested
        for p in self.model.parameters():
            p.requires_grad_(trainable)

    def forward(self, inputs: torch.Tensor) -> Dict[str, torch.Tensor]:
        return {"encoder_output": self.model(inputs)}

    @abstractmethod
    def _remove_softmax_layer(self):
        """Model specific method that allows the final softmax layer to be implemented in the Ludwig Decoder
        component.  The model specific implementation should change the final softmax layer in the torchvision
        model architecture to torch.nn.Identity().  This allows the output tensor from the preceding layer to be
        passed to the Ludwig Combiner and then to the Decoder.

        Returns: None
        """
        raise NotImplementedError()

    @property
    def output_shape(self) -> torch.Size:
        # create synthetic image and run through forward method
        inputs = torch.randn([1, *self.input_shape])
        output = self.model(inputs)
        return torch.Size(output.shape[1:])

    @property
    def input_shape(self) -> torch.Size:
        # expected shape after all pre-processing
        # len(transforms_obj.mean) determines the number of channels
        # transforms_obj.crop_size determines the height and width of image
        # [num_channels, height, width]
        return torch.Size([self.num_channels, *(2 * self.crop_size)])


@DeveloperAPI
@register_torchvision_model_variants(
    [
        TVModelVariant(variant_id="base", create_model_function=tvm.alexnet, model_weights=tvm.AlexNet_Weights),
    ]
)
@register_encoder("alexnet", IMAGE)
class TVAlexNetEncoder(TVBaseEncoder):
    # specify base model type
    torchvision_model_type: str = "alexnet"

    def __init__(
        self,
        **kwargs,
    ):
        logger.debug(f" {self.name}")
        super().__init__(**kwargs)

    # TODO: discussion w/ justin
    # @property
    # def get_torchvision_model_type(self):
    #     return "alexnet"

    def _remove_softmax_layer(self):
        self.model.classifier[-1] = torch.nn.Identity()

    @staticmethod
    def get_schema_cls():
        return TVAlexNetEncoderConfig


@DeveloperAPI
@register_torchvision_model_variants(
    [
        TVModelVariant(
            variant_id="tiny", create_model_function=tvm.convnext_tiny, model_weights=tvm.ConvNeXt_Tiny_Weights
        ),
        TVModelVariant(
            variant_id="small", create_model_function=tvm.convnext_small, model_weights=tvm.ConvNeXt_Small_Weights
        ),
        TVModelVariant(
            variant_id="base", create_model_function=tvm.convnext_base, model_weights=tvm.ConvNeXt_Base_Weights
        ),
        TVModelVariant(
            variant_id="large", create_model_function=tvm.convnext_large, model_weights=tvm.ConvNeXt_Large_Weights
        ),
    ]
)
@register_encoder("convnext", IMAGE)
class TVConvNeXtEncoder(TVBaseEncoder):
    # specify base torchvision model
    torchvision_model_type: str = "convnext"

    def __init__(
        self,
        **kwargs,
    ):
        logger.debug(f" {self.name}")
        super().__init__(**kwargs)

    def _remove_softmax_layer(self):
        self.model.classifier[-1] = torch.nn.Identity()

    @staticmethod
    def get_schema_cls():
        return TVConvNeXtEncoderConfig


@DeveloperAPI
@register_torchvision_model_variants(
    [
        TVModelVariant(121, tvm.densenet121, tvm.DenseNet121_Weights),
        TVModelVariant(161, tvm.densenet161, tvm.DenseNet161_Weights),
        TVModelVariant(169, tvm.densenet169, tvm.DenseNet169_Weights),
        TVModelVariant(201, tvm.densenet201, tvm.DenseNet201_Weights),
    ]
)
@register_encoder("densenet", IMAGE)
class TVDenseNetEncoder(TVBaseEncoder):
    # specify base torchvision model
    torchvision_model_type: str = "densenet"

    def __init__(
        self,
        **kwargs,
    ):
        logger.debug(f" {self.name}")
        super().__init__(**kwargs)

    def _remove_softmax_layer(self):
        self.model.classifier = torch.nn.Identity()

    @staticmethod
    def get_schema_cls():
        return TVDenseNetEncoderConfig


@DeveloperAPI
@register_torchvision_model_variants(
    [
        TVModelVariant("b0", tvm.efficientnet_b0, tvm.EfficientNet_B0_Weights),
        TVModelVariant("b1", tvm.efficientnet_b1, tvm.EfficientNet_B1_Weights),
        TVModelVariant("b2", tvm.efficientnet_b2, tvm.EfficientNet_B2_Weights),
        TVModelVariant("b3", tvm.efficientnet_b3, tvm.EfficientNet_B3_Weights),
        TVModelVariant("b4", tvm.efficientnet_b4, tvm.EfficientNet_B4_Weights),
        TVModelVariant("b5", tvm.efficientnet_b5, tvm.EfficientNet_B5_Weights),
        TVModelVariant("b6", tvm.efficientnet_b6, tvm.EfficientNet_B6_Weights),
        TVModelVariant("b7", tvm.efficientnet_b7, tvm.EfficientNet_B7_Weights),
        TVModelVariant("v2_s", tvm.efficientnet_v2_s, tvm.EfficientNet_V2_S_Weights),
        TVModelVariant("v2_m", tvm.efficientnet_v2_m, tvm.EfficientNet_V2_M_Weights),
        TVModelVariant("v2_l", tvm.efficientnet_v2_l, tvm.EfficientNet_V2_L_Weights),
    ]
)
@register_encoder("efficientnet", IMAGE)
class TVEfficientNetEncoder(TVBaseEncoder):
    # specify base torchvision model
    torchvision_model_type: str = "efficientnet"

    def __init__(
        self,
        **kwargs,
    ):
        logger.debug(f" {self.name}")
        super().__init__(**kwargs)

    def _remove_softmax_layer(self):
        self.model.classifier[-1] = torch.nn.Identity()

    @staticmethod
    def get_schema_cls():
        return TVEfficientNetEncoderConfig


@DeveloperAPI
@register_torchvision_model_variants(
    [
        TVModelVariant("base", tvm.googlenet, tvm.GoogLeNet_Weights),
    ]
)
@register_encoder("googlenet", IMAGE)
class TVGoogLeNetEncoder(TVBaseEncoder):
    # specify base torchvision model
    torchvision_model_type: str = "googlenet"

    def __init__(
        self,
        **kwargs,
    ):
        logger.debug(f" {self.name}")
        super().__init__(**kwargs)

        # if auxiliary network exists, eliminate auxiliary network
        # to resolve issue when loading a saved model which does not
        # contain the auxiliary network
        if self.model.aux_logits:
            self.model.aux_logits = False
            self.model.aux1 = None
            self.model.aux2 = None

    def _remove_softmax_layer(self):
        self.model.fc = torch.nn.Identity()

    @staticmethod
    def get_schema_cls():
        return TVGoogLeNetEncoderConfig


@DeveloperAPI
@register_torchvision_model_variants(
    [
        TVModelVariant("base", tvm.inception_v3, tvm.Inception_V3_Weights),
    ]
)
@register_encoder("inceptionv3", IMAGE)
class TVInceptionV3Encoder(TVBaseEncoder):
    # specify base torchvision model
    torchvision_model_type: str = "inceptionv3"

    def __init__(
        self,
        **kwargs,
    ):
        logger.debug(f" {self.name}")
        super().__init__(**kwargs)

        # if auxiliary network exists, eliminate auxiliary network
        # to resolve issue when loading a saved model which does not
        # contain the auxiliary network
        if self.model.aux_logits:
            self.model.aux_logits = False
            self.model.AuxLogits = None

    def _remove_softmax_layer(self):
        self.model.fc = torch.nn.Identity()

    @staticmethod
    def get_schema_cls():
        return TVInceptionV3EncoderConfig


@DeveloperAPI
@register_torchvision_model_variants(
    [
        TVModelVariant("t", tvm.maxvit_t, tvm.MaxVit_T_Weights),
    ]
)
@register_encoder("maxvit", IMAGE)
class TVMaxVitEncoder(TVBaseEncoder):
    # specify base torchvision model
    torchvision_model_type: str = "maxvit"

    def __init__(
        self,
        **kwargs,
    ):
        logger.debug(f" {self.name}")
        super().__init__(**kwargs)

    def _remove_softmax_layer(self):
        self.model.classifier[-1] = torch.nn.Identity()

    @staticmethod
    def get_schema_cls():
        return TVMaxVitEncoderConfig


@DeveloperAPI
@register_torchvision_model_variants(
    [
        TVModelVariant("0_5", tvm.mnasnet0_5, tvm.mnasnet.MNASNet0_5_Weights),
        TVModelVariant("0_75", tvm.mnasnet0_75, tvm.mnasnet.MNASNet0_75_Weights),
        TVModelVariant("1_0", tvm.mnasnet1_0, tvm.mnasnet.MNASNet1_0_Weights),
        TVModelVariant("1_3", tvm.mnasnet1_3, tvm.mnasnet.MNASNet1_3_Weights),
    ]
)
@register_encoder("mnasnet", IMAGE)
class TVMNASNetEncoder(TVBaseEncoder):
    # specify base torchvision model
    torchvision_model_type: str = "mnasnet"

    def __init__(
        self,
        **kwargs,
    ):
        logger.debug(f" {self.name}")
        super().__init__(**kwargs)

    def _remove_softmax_layer(self):
        self.model.classifier[-1] = torch.nn.Identity()

    @staticmethod
    def get_schema_cls():
        return TVMNASNetEncoderConfig


@DeveloperAPI
@register_torchvision_model_variants(
    [
        TVModelVariant("base", tvm.mobilenet_v2, tvm.MobileNet_V2_Weights),
    ]
)
@register_encoder("mobilenetv2", IMAGE)
class TVMobileNetV2Encoder(TVBaseEncoder):
    # specify base torchvision model
    torchvision_model_type: str = "mobilenetv2"

    def __init__(
        self,
        **kwargs,
    ):
        logger.debug(f" {self.name}")
        super().__init__(**kwargs)

    def _remove_softmax_layer(self):
        self.model.classifier[-1] = torch.nn.Identity()

    @staticmethod
    def get_schema_cls():
        return TVMobileNetV2EncoderConfig


@DeveloperAPI
@register_torchvision_model_variants(
    [
        TVModelVariant("small", tvm.mobilenet_v3_small, tvm.MobileNet_V3_Small_Weights),
        TVModelVariant("large", tvm.mobilenet_v3_large, tvm.MobileNet_V3_Large_Weights),
    ]
)
@register_encoder("mobilenetv3", IMAGE)
class TVMobileNetV3Encoder(TVBaseEncoder):
    # specify base torchvision model
    torchvision_model_type: str = "mobilenetv3"

    def __init__(
        self,
        **kwargs,
    ):
        logger.debug(f" {self.name}")
        super().__init__(**kwargs)

    def _remove_softmax_layer(self):
        self.model.classifier[-1] = torch.nn.Identity()

    @staticmethod
    def get_schema_cls():
        return TVMobileNetV3EncoderConfig


@DeveloperAPI
@register_torchvision_model_variants(
    [
        TVModelVariant("x_16gf", tvm.regnet_x_16gf, tvm.RegNet_X_16GF_Weights),
        TVModelVariant("x_1_6gf", tvm.regnet_x_1_6gf, tvm.RegNet_X_1_6GF_Weights),
        TVModelVariant("x_32gf", tvm.regnet_x_32gf, tvm.RegNet_X_32GF_Weights),
        TVModelVariant("x_3_2gf", tvm.regnet_x_3_2gf, tvm.RegNet_X_3_2GF_Weights),
        TVModelVariant("x_400mf", tvm.regnet_x_400mf, tvm.RegNet_X_400MF_Weights),
        TVModelVariant("x_800mf", tvm.regnet_x_800mf, tvm.RegNet_X_800MF_Weights),
        TVModelVariant("x_8gf", tvm.regnet_x_8gf, tvm.RegNet_X_8GF_Weights),
        TVModelVariant("y_128gf", tvm.regnet_y_128gf, tvm.RegNet_Y_128GF_Weights),
        TVModelVariant("y_16gf", tvm.regnet_y_16gf, tvm.RegNet_Y_16GF_Weights),
        TVModelVariant("y_1_6gf", tvm.regnet_y_1_6gf, tvm.RegNet_Y_1_6GF_Weights),
        TVModelVariant("y_32gf", tvm.regnet_y_32gf, tvm.RegNet_Y_32GF_Weights),
        TVModelVariant("y_3_2gf", tvm.regnet_y_3_2gf, tvm.RegNet_Y_3_2GF_Weights),
        TVModelVariant("y_400mf", tvm.regnet_y_400mf, tvm.RegNet_Y_400MF_Weights),
        TVModelVariant("y_800mf", tvm.regnet_y_800mf, tvm.RegNet_Y_800MF_Weights),
        TVModelVariant("y_8gf", tvm.regnet_y_8gf, tvm.RegNet_Y_8GF_Weights),
    ]
)
@register_encoder("regnet", IMAGE)
class TVRegNetEncoder(TVBaseEncoder):
    # specify base torchvision model
    torchvision_model_type: str = "regnet"

    def __init__(
        self,
        **kwargs,
    ):
        logger.debug(f" {self.name}")
        super().__init__(**kwargs)

    def _remove_softmax_layer(self):
        self.model.fc = torch.nn.Identity()

    @staticmethod
    def get_schema_cls():
        return TVRegNetEncoderConfig


@DeveloperAPI
@register_torchvision_model_variants(
    [
        TVModelVariant(18, tvm.resnet18, tvm.ResNet18_Weights),
        TVModelVariant(34, tvm.resnet34, tvm.ResNet34_Weights),
        TVModelVariant(50, tvm.resnet50, tvm.ResNet50_Weights),
        TVModelVariant(101, tvm.resnet101, tvm.ResNet101_Weights),
        TVModelVariant(152, tvm.resnet152, tvm.ResNet152_Weights),
    ]
)
@register_encoder("resnet", IMAGE)
class TVResNetEncoder(TVBaseEncoder):
    # specify base torchvision model
    torchvision_model_type: str = "resnet"

    def __init__(
        self,
        **kwargs,
    ):
        logger.debug(f" {self.name}")
        super().__init__(**kwargs)

    def _remove_softmax_layer(self):
        self.model.fc = torch.nn.Identity()

    @staticmethod
    def get_schema_cls():
        return TVResNetEncoderConfig


@DeveloperAPI
@register_torchvision_model_variants(
    [
        TVModelVariant("50_32x4d", tvm.resnext50_32x4d, tvm.ResNeXt50_32X4D_Weights),
        TVModelVariant("101_328xd", tvm.resnext101_32x8d, tvm.ResNeXt101_32X8D_Weights),
        TVModelVariant("101_64x4d", tvm.resnext101_64x4d, tvm.ResNeXt101_64X4D_Weights),
    ]
)
@register_encoder("resnext", IMAGE)
class TVResNeXtEncoder(TVBaseEncoder):
    # specify base torchvision model
    torchvision_model_type: str = "resnext"

    def __init__(
        self,
        **kwargs,
    ):
        logger.debug(f" {self.name}")
        super().__init__(**kwargs)

    def _remove_softmax_layer(self):
        self.model.fc = torch.nn.Identity()

    @staticmethod
    def get_schema_cls():
        return TVResNeXtEncoderConfig


@DeveloperAPI
@register_torchvision_model_variants(
    [
        TVModelVariant("x0_5", tvm.shufflenet_v2_x0_5, tvm.ShuffleNet_V2_X0_5_Weights),
        TVModelVariant("x1_0", tvm.shufflenet_v2_x1_0, tvm.ShuffleNet_V2_X1_0_Weights),
        TVModelVariant("x1_5", tvm.shufflenet_v2_x1_5, tvm.ShuffleNet_V2_X1_5_Weights),
        TVModelVariant("x2_0", tvm.shufflenet_v2_x2_0, tvm.ShuffleNet_V2_X2_0_Weights),
    ]
)
@register_encoder("shufflenet_v2", IMAGE)
class TVShuffleNetV2Encoder(TVBaseEncoder):
    # specify base torchvision model
    torchvision_model_type: str = "shufflenet_v2"

    def __init__(
        self,
        **kwargs,
    ):
        logger.debug(f" {self.name}")
        super().__init__(**kwargs)

    def _remove_softmax_layer(self):
        self.model.fc = torch.nn.Identity()

    @staticmethod
    def get_schema_cls():
        return TVShuffleNetV2EncoderConfig


@DeveloperAPI
@register_torchvision_model_variants(
    [
        TVModelVariant("1_0", tvm.squeezenet1_0, tvm.SqueezeNet1_0_Weights),
        TVModelVariant("1_1", tvm.squeezenet1_1, tvm.SqueezeNet1_1_Weights),
    ]
)
@register_encoder("squeezenet", IMAGE)
class TVSqueezeNetEncoder(TVBaseEncoder):
    # specify base torchvision model
    torchvision_model_type: str = "squeezenet"

    def __init__(
        self,
        **kwargs,
    ):
        logger.debug(f" {self.name}")
        super().__init__(**kwargs)

    def _remove_softmax_layer(self):
        # SqueezeNet does not have a final nn.Linear() layer
        # Use flatten output from last AdaptiveAvgPool2d layer
        # as encoder output.
        pass

    @staticmethod
    def get_schema_cls():
        return TVSqueezeNetEncoderConfig


@DeveloperAPI
@register_torchvision_model_variants(
    [
        TVModelVariant("t", tvm.swin_t, tvm.Swin_T_Weights),
        TVModelVariant("s", tvm.swin_s, tvm.Swin_S_Weights),
        TVModelVariant("b", tvm.swin_b, tvm.Swin_B_Weights),
    ]
)
@register_encoder("swin_transformer", IMAGE)
class TVSwinTransformerEncoder(TVBaseEncoder):
    # specify base torchvision model
    torchvision_model_type: str = "swin_transformer"

    def __init__(
        self,
        **kwargs,
    ):
        logger.debug(f" {self.name}")
        super().__init__(**kwargs)

    def _remove_softmax_layer(self):
        self.model.head = torch.nn.Identity()

    @staticmethod
    def get_schema_cls():
        return TVSwinTransformerEncoderConfig


@DeveloperAPI
@register_torchvision_model_variants(
    [
        TVModelVariant(11, tvm.vgg11, tvm.VGG11_Weights),
        TVModelVariant("11_bn", tvm.vgg11_bn, tvm.VGG11_BN_Weights),
        TVModelVariant(13, tvm.vgg13, tvm.VGG13_Weights),
        TVModelVariant("13_bn", tvm.vgg13_bn, tvm.VGG13_BN_Weights),
        TVModelVariant(16, tvm.vgg16, tvm.VGG16_Weights),
        TVModelVariant("16_bn", tvm.vgg16_bn, tvm.VGG16_BN_Weights),
        TVModelVariant(19, tvm.vgg19, tvm.VGG19_Weights),
        TVModelVariant("19_bn", tvm.vgg19_bn, tvm.VGG19_BN_Weights),
    ]
)
@register_encoder("vgg", IMAGE)
class TVVGGEncoder(TVBaseEncoder):
    # specify base torchvison model
    torchvision_model_type: str = "vgg"

    def __init__(
        self,
        **kwargs,
    ):
        logger.debug(f" {self.name}")
        super().__init__(**kwargs)

    def _remove_softmax_layer(self):
        self.model.classifier[-1] = torch.nn.Identity()

    @staticmethod
    def get_schema_cls():
        return TVVGGEncoderConfig


@DeveloperAPI
@register_torchvision_model_variants(
    [
        TVModelVariant("b_16", tvm.vit_b_16, tvm.ViT_B_16_Weights),
        TVModelVariant("b_32", tvm.vit_b_32, tvm.ViT_B_32_Weights),
        TVModelVariant("l_16", tvm.vit_l_16, tvm.ViT_L_16_Weights),
        TVModelVariant("l_32", tvm.vit_l_32, tvm.ViT_L_32_Weights),
        TVModelVariant("h_14", tvm.vit_h_14, tvm.ViT_H_14_Weights),
    ]
)
@register_encoder("vit", IMAGE)
class TVViTEncoder(TVBaseEncoder):
    # specify base torchvision model
    torchvision_model_type: str = "vit"

    def __init__(
        self,
        **kwargs,
    ):
        logger.debug(f" {self.name}")

        # Depending on model variant and weight specification, the expected image size
        # will vary.  This code determines at run time what the expected image size will be
        # and adds to the kwargs dictionary the parameter that specifies the image size.
        # this is needed only if not using pretrained weights.  If pre-trained weights are
        # specified, then the correct image size is set.
        if not kwargs["use_pretrained"]:
            weights_specification = torchvision_model_registry[self.torchvision_model_type][
                kwargs["model_variant"]
            ].model_weights.DEFAULT
            kwargs["image_size"] = weights_specification.transforms.keywords["crop_size"]

        super().__init__(**kwargs)

    def _remove_softmax_layer(self):
        self.model.heads[-1] = torch.nn.Identity()

    @staticmethod
    def get_schema_cls():
        return TVViTEncoderConfig


@DeveloperAPI
@register_torchvision_model_variants(
    [
        TVModelVariant("50_2", tvm.wide_resnet50_2, tvm.Wide_ResNet50_2_Weights),
        TVModelVariant("101_2", tvm.wide_resnet101_2, tvm.Wide_ResNet101_2_Weights),
    ]
)
@register_encoder("wide_resnet", IMAGE)
class TVWideResNetEncoder(TVBaseEncoder):
    # specify base torchvision model
    torchvision_model_type: str = "wide_resnet"

    def __init__(
        self,
        **kwargs,
    ):
        logger.debug(f" {self.name}")
        super().__init__(**kwargs)

    def _remove_softmax_layer(self):
        self.model.fc = torch.nn.Identity()

    @staticmethod
    def get_schema_cls():
        return TVWideResNetEncoderConfig



import os
import tempfile
from typing import Dict

import torch
from marshmallow_dataclass import dataclass
from torch import nn, Tensor

from ludwig.api import LudwigModel
from ludwig.combiners.combiners import Combiner, register_combiner
from ludwig.constants import BATCH_SIZE, LOGITS, MINIMIZE, NUMBER, TRAINER
from ludwig.decoders.base import Decoder
from ludwig.decoders.registry import register_decoder
from ludwig.encoders.base import Encoder
from ludwig.encoders.registry import register_encoder
from ludwig.modules.loss_modules import LogitsInputsMixin, register_loss
from ludwig.modules.metric_modules import LossMetric, register_metric
from ludwig.schema import utils as schema_utils
from ludwig.schema.combiners.base import BaseCombinerConfig
from ludwig.schema.decoders.base import BaseDecoderConfig
from ludwig.schema.decoders.utils import register_decoder_config
from ludwig.schema.encoders.base import BaseEncoderConfig
from ludwig.schema.encoders.utils import register_encoder_config
from ludwig.schema.features.loss.loss import BaseLossConfig
from ludwig.schema.features.loss.loss import register_loss as register_loss_schema
from tests.integration_tests.utils import (
    category_feature,
    generate_data,
    LocalTestBackend,
    number_feature,
    sequence_feature,
)


@dataclass
class CustomTestCombinerConfig(BaseCombinerConfig):
    type: str = "custom_combiner"

    foo: bool = schema_utils.Boolean(default=False, description="")


@register_encoder_config("custom_number_encoder", NUMBER)
@dataclass
class CustomNumberEncoderConfig(BaseEncoderConfig):
    type: str = "custom_number_encoder"

    input_size: int = schema_utils.PositiveInteger(default=1, description="")


@register_decoder_config("custom_number_decoder", NUMBER)
@dataclass
class CustomNumberDecoderConfig(BaseDecoderConfig):
    type: str = "custom_number_decoder"

    input_size: int = schema_utils.PositiveInteger(default=1, description="")


@register_loss_schema([NUMBER])
@dataclass
class CustomLossConfig(BaseLossConfig):
    type: str = "custom_loss"


@register_combiner(name="custom_test")
class CustomTestCombiner(Combiner):
    def __init__(self, input_features: Dict = None, config: CustomTestCombinerConfig = None, **kwargs):
        super().__init__(input_features)
        self.foo = config.foo

    def forward(self, inputs: Dict) -> Dict:  # encoder outputs
        if not self.foo:
            raise ValueError("expected foo to be True")

        # minimal transformation from inputs to outputs
        encoder_outputs = [inputs[k]["encoder_output"] for k in inputs]
        hidden = torch.cat(encoder_outputs, 1)
        return_data = {"combiner_output": hidden}

        return return_data

    @staticmethod
    def get_schema_cls():
        return CustomTestCombinerConfig


@register_encoder("custom_number_encoder", NUMBER)
class CustomNumberEncoder(Encoder):
    def __init__(self, input_size, **kwargs):
        super().__init__()
        self.input_size = input_size

    def forward(self, inputs, **kwargs):
        return {"encoder_output": inputs}

    @property
    def input_shape(self) -> torch.Size:
        return torch.Size([self.input_size])

    @property
    def output_shape(self) -> torch.Size:
        return self.input_shape

    @staticmethod
    def get_schema_cls():
        return CustomNumberEncoderConfig


@register_decoder("custom_number_decoder", NUMBER)
class CustomNumberDecoder(Decoder):
    def __init__(self, input_size, **kwargs):
        super().__init__()
        self.input_size = input_size

    @property
    def input_shape(self):
        return torch.Size([self.input_size])

    def forward(self, inputs, **kwargs):
        return torch.mean(inputs, 1)

    @staticmethod
    def get_schema_cls():
        return CustomNumberDecoderConfig


@register_loss("custom_loss", [NUMBER])
class CustomLoss(nn.Module, LogitsInputsMixin):
    def __init__(self, **kwargs):
        super().__init__()

    def forward(self, preds: Tensor, target: Tensor) -> Tensor:
        return torch.mean(torch.square(preds - target))

    @staticmethod
    def get_schema_cls():
        return CustomLossConfig


@register_metric("custom_loss", [NUMBER], MINIMIZE, LOGITS)
class CustomLossMetric(LossMetric):
    def __init__(self, **kwargs):
        super().__init__()
        self.loss_fn = CustomLoss()

    def get_current_value(self, preds: Tensor, target: Tensor):
        return self.loss_fn(preds, target)


def test_custom_combiner():
    _run_test(combiner={"type": "custom_test", "foo": True})


def test_custom_encoder_decoder():
    input_features = [
        sequence_feature(encoder={"reduce_output": "sum"}),
        number_feature(encoder={"type": "custom_number_encoder"}),
    ]
    output_features = [
        number_feature(decoder={"type": "custom_number_decoder"}),
    ]
    _run_test(input_features=input_features, output_features=output_features)


def test_custom_loss_metric():
    output_features = [
        number_feature(loss={"type": "custom_loss"}),
    ]
    _run_test(output_features=output_features)


def _run_test(input_features=None, output_features=None, combiner=None):
    with tempfile.TemporaryDirectory() as tmpdir:
        input_features = input_features or [
            sequence_feature(encoder={"reduce_output": "sum"}),
            number_feature(),
        ]
        output_features = output_features or [category_feature(decoder={"vocab_size": 2}, reduce_input="sum")]
        combiner = combiner or {"type": "concat"}

        csv_filename = os.path.join(tmpdir, "training.csv")
        data_csv = generate_data(input_features, output_features, csv_filename)

        config = {
            "input_features": input_features,
            "output_features": output_features,
            "combiner": combiner,
            TRAINER: {"epochs": 2, BATCH_SIZE: 128},
        }

        model = LudwigModel(config, backend=LocalTestBackend())
        _, _, output_directory = model.train(
            dataset=data_csv,
            output_directory=tmpdir,
        )
        model.predict(dataset=data_csv, output_directory=output_directory)



from collections import Counter, defaultdict
from statistics import mean
from typing import Any, Dict, List, Tuple, Union

import torch
from torch._C._autograd import _KinetoEvent
from torch.autograd import DeviceType, profiler_util

from ludwig.benchmarking.profiler_dataclasses import DeviceUsageMetrics, SystemResourceMetrics, TorchProfilerMetrics
from ludwig.constants import LUDWIG_TAG


def initialize_stats_dict(main_function_events: List[profiler_util.FunctionEvent]) -> Dict[str, List]:
    """Initialize dictionary which stores resource usage information per tagged code block.

    :param main_function_events: list of main function events.
    """
    info = {}
    for event_name in [evt.name for evt in main_function_events]:
        info[event_name] = []
    return info


def get_memory_details(kineto_event: _KinetoEvent) -> Tuple[str, int]:
    """Get device name and number of bytes (de)allocated during an event.

    :param kineto_event: a Kineto event instance.
    """
    if kineto_event.device_type() in [DeviceType.CPU, DeviceType.MKLDNN, DeviceType.IDEEP]:
        return "cpu", kineto_event.nbytes()
    elif kineto_event.device_type() in [DeviceType.CUDA, DeviceType.HIP]:
        return f"cuda_{kineto_event.device_index()}", kineto_event.nbytes()
    else:
        raise ValueError(f"Device {kineto_event.device_type()} is not valid.")


def get_device_memory_usage(
    kineto_event: _KinetoEvent, memory_events: List[List[Union[_KinetoEvent, bool]]]
) -> Dict[str, DeviceUsageMetrics]:
    """Get CPU and CUDA memory usage for an event.

    :param kineto_event: a Kineto event instance.
    :param memory_events: list of memory events.
    """
    mem_records_acc = profiler_util.MemRecordsAcc(memory_events)
    records_in_interval = mem_records_acc.in_interval(
        kineto_event.start_us(), kineto_event.start_us() + kineto_event.duration_us()
    )
    memory_so_far = defaultdict(int)
    count_so_far = defaultdict(int)
    average_so_far = defaultdict(float)
    max_so_far = defaultdict(int)

    for mem_record in records_in_interval:
        device, nbytes = get_memory_details(mem_record[0])
        memory_so_far[device] += nbytes
        max_so_far[device] = max(max_so_far[device], memory_so_far[device])
        average_so_far[device] = (memory_so_far[device] + (average_so_far[device] * count_so_far[device])) / (
            count_so_far[device] + 1
        )
        count_so_far[device] += 1
    memory_info_per_device = {}
    for device in count_so_far:
        memory_info_per_device[f"torch_{device}_"] = DeviceUsageMetrics(
            max_memory_used=max_so_far[device], average_memory_used=average_so_far[device]
        )
    return memory_info_per_device


def get_torch_op_time(events: List[profiler_util.FunctionEvent], attr: str) -> Union[int, float]:
    """Get time torch operators spent executing for a list of events.

    :param events: list of events.
    :param attr: a FunctionEvent attribute. Expecting one of "cpu_time_total", "cuda_time_total".
    """
    if attr not in ["cpu_time_total", "cuda_time_total"]:
        return -1

    total = 0
    for e in events:
        # Possible trace_names are torch ops, or tagged code blocks by LudwigProfiler (which are
        # prepended with LUDWIG_TAG).
        if LUDWIG_TAG not in e.trace_name:
            total += getattr(e, attr)
        else:
            total += get_torch_op_time(e.cpu_children, attr)
    return total


def get_device_run_durations(function_event: profiler_util.FunctionEvent) -> Tuple[float, float]:
    """Get CPU and CUDA run durations for an event.

    :param function_event: a function event instance.
    """
    torch_cpu_time = get_torch_op_time(function_event.cpu_children, "cpu_time_total")
    torch_cuda_time = get_torch_op_time(function_event.cpu_children, "cuda_time_total")
    return torch_cpu_time, torch_cuda_time


def get_num_oom_events(kineto_event: _KinetoEvent, out_of_memory_events: List[List[Union[_KinetoEvent, bool]]]) -> int:
    oom_records_acc = profiler_util.MemRecordsAcc(out_of_memory_events)
    records_in_interval = oom_records_acc.in_interval(
        kineto_event.start_us(), kineto_event.start_us() + kineto_event.duration_us()
    )
    return len(list(records_in_interval))


def get_resource_usage_report(
    main_kineto_events: List[_KinetoEvent],
    main_function_events: List[profiler_util.FunctionEvent],
    memory_events: List[List[Union[_KinetoEvent, bool]]],
    out_of_memory_events: List[List[Union[_KinetoEvent, bool]]],
    info: Dict[str, Any],
) -> Dict[str, List[TorchProfilerMetrics]]:
    """Get relevant information from Kineto events and function events exported by the profiler.

    :param main_kineto_events: list of main Kineto events.
    :param main_function_events: list of main function events.
    :param memory_events: list of memory events.
    :param out_of_memory_events: list of out of memory events.
    :param info: dictionary used to record resource usage metrics.
    """
    main_kineto_events = sorted(
        (evt for evt in main_kineto_events if LUDWIG_TAG in evt.name()), key=lambda x: x.correlation_id()
    )
    main_function_events = sorted((evt for evt in main_function_events if LUDWIG_TAG in evt.name), key=lambda x: x.id)

    for kineto_event, function_event in zip(main_kineto_events, main_function_events):
        # Two different instances of `function_event` can have the same name if a the same
        # tagged code block/function was executed more than once.
        memory_info_per_device = get_device_memory_usage(kineto_event, memory_events)
        torch_cpu_time, torch_cuda_time = get_device_run_durations(function_event)
        num_oom_events = get_num_oom_events(kineto_event, out_of_memory_events)
        torch_profiler_metrics = TorchProfilerMetrics(
            torch_cpu_time=torch_cpu_time,
            torch_cuda_time=torch_cuda_time,
            num_oom_events=num_oom_events,
            device_usage=memory_info_per_device,
        )
        info[function_event.name].append(torch_profiler_metrics)
    return info


def get_all_events(
    kineto_events: List[_KinetoEvent], function_events: profiler_util.EventList
) -> Tuple[
    List[_KinetoEvent],
    List[profiler_util.FunctionEvent],
    List[List[Union[_KinetoEvent, bool]]],
    List[List[Union[_KinetoEvent, bool]]],
]:
    """Return main Kineto and function events, memory and OOM events for functions/code blocks tagged in
    LudwigProfiler.

    :param kineto_events: list of Kineto Events.
    :param function_events: list of function events.
    """
    # LUDWIG_TAG is prepended to LudwigProfiler tags. This edited tag is passed in to `torch.profiler.record_function`
    # so we can easily retrieve events for code blocks wrapped with LudwigProfiler.
    main_function_events = [evt for evt in function_events if LUDWIG_TAG in evt.name]
    main_kineto_events = [event for event in kineto_events if LUDWIG_TAG in event.name()]
    memory_events = [[event, False] for event in kineto_events if profiler_util.MEMORY_EVENT_NAME in event.name()]
    # profiler_util.OUT_OF_MEMORY_EVENT_NAME seems to only be in newer versions of torch.
    out_of_memory_events = [[event, False] for event in kineto_events if "[OutOfMemory]" in event.name()]
    return main_kineto_events, main_function_events, memory_events, out_of_memory_events


def get_metrics_from_torch_profiler(profile: torch.profiler.profiler.profile) -> Dict[str, List[TorchProfilerMetrics]]:
    """Export time and resource usage metrics (CPU and CUDA) from a PyTorch profiler.

    The profiler keeps track of *torch operations* being executed in C++. It keeps track
    of what device they're executed on, their execution time, and memory usage.
    We only track the aforementioned metrics, but the torch profiler can keep track of
    the stack trace, FLOPs, and torch modules. Tracking each additional item adds overhead.

    The torch profiler surfaces these metrics that are tracked under the hood by `libkineto`.
    More on the Kineto project: https://github.com/pytorch/kineto

    :param profile: profiler object that contains all the events that
        were registered during the execution of the wrapped code block.
    """
    # events in both of these lists are in chronological order.
    kineto_events = profile.profiler.kineto_results.events()
    function_events = profile.profiler.function_events
    main_kineto_events, main_function_events, memory_events, out_of_memory_events = get_all_events(
        kineto_events, function_events
    )

    assert Counter([event.name for event in main_function_events]) == Counter(
        [event.name() for event in main_kineto_events]
    )
    info = initialize_stats_dict(main_function_events)
    info = get_resource_usage_report(
        main_kineto_events, main_function_events, memory_events, out_of_memory_events, info
    )
    return info


def get_metrics_from_system_usage_profiler(system_usage_info: dict) -> SystemResourceMetrics:
    """Package system resource usage metrics (no torch operators) in a dataclass.

    :param system_usage_info: dictionary containing resource usage information.
    """
    device_usage_dict: Dict[str, DeviceUsageMetrics] = {}
    for key in system_usage_info:
        if "cuda_" in key and "_memory_used" in key:
            cuda_device_name = "_".join(key.split("_")[:2]) + "_"
            max_memory_used = max(system_usage_info[key], default=0)
            average_memory_used = mean(system_usage_info.get(key, [0]))
            device_usage_dict[cuda_device_name] = DeviceUsageMetrics(
                max_memory_used=max_memory_used, average_memory_used=average_memory_used
            )
    return SystemResourceMetrics(
        code_block_tag=system_usage_info["code_block_tag"],
        cpu_name=system_usage_info.get("cpu_name", "unknown"),
        cpu_architecture=system_usage_info["cpu_architecture"],
        num_cpu=system_usage_info["num_cpu"],
        total_cpu_memory_size=system_usage_info["total_cpu_memory_size"],
        ludwig_version=system_usage_info["ludwig_version"],
        total_execution_time=system_usage_info["end_time"] - system_usage_info["start_time"],
        disk_footprint=system_usage_info["end_disk_usage"] - system_usage_info["start_disk_usage"],
        max_cpu_utilization=max(system_usage_info["cpu_utilization"], default=0),
        max_cpu_memory_usage=max(system_usage_info["cpu_memory_usage"], default=0),
        min_global_cpu_memory_available=min(system_usage_info["global_cpu_memory_available"], default=0),
        max_global_cpu_utilization=max(system_usage_info["global_cpu_utilization"], default=0),
        average_cpu_utilization=mean(system_usage_info.get("cpu_utilization", [0])),
        average_cpu_memory_usage=mean(system_usage_info.get("cpu_memory_usage", [0])),
        average_global_cpu_memory_available=mean(system_usage_info.get("global_cpu_memory_available", [0])),
        average_global_cpu_utilization=mean(system_usage_info.get("global_cpu_utilization", [0])),
        device_usage=device_usage_dict,
    )



import logging
from typing import Set

from ludwig.api_annotations import DeveloperAPI
from ludwig.constants import AUDIO, BINARY, CATEGORY, DATE, IMAGE, NUMBER, TEXT
from ludwig.utils import strings_utils
from ludwig.utils.automl.field_info import FieldInfo

# For a given feature, the highest percentage of distinct values out of the total number of rows that we might still
# assign the CATEGORY type.
CATEGORY_TYPE_DISTINCT_VALUE_PERCENTAGE_CUTOFF = 0.5

# Consider the field a valid text field if it has at least 5 average words. Fewer than this and it may be a cateogry
# or an ID field (like a name or place) of some kind.
TEXT_AVG_WORDS_CUTOFF = 5


@DeveloperAPI
def infer_type(field: FieldInfo, missing_value_percent: float, row_count: int) -> str:
    """Perform type inference on field.

    # Inputs
    :param field: (FieldInfo) object describing field
    :param missing_value_percent: (float) percent of missing values in the column
    :param row_count: (int) total number of entries in original dataset

    # Return
    :return: (str) feature type
    """
    if field.dtype == DATE:
        return DATE

    num_distinct_values = field.num_distinct_values
    distinct_values = field.distinct_values

    if num_distinct_values <= 1:
        return CATEGORY

    if num_distinct_values == 2 and missing_value_percent == 0:
        # Check that all distinct values are conventional bools.
        if strings_utils.are_conventional_bools(distinct_values):
            return BINARY

    if field.image_values >= 3:
        return IMAGE

    if field.audio_values >= 3:
        return AUDIO

    # Use CATEGORY if:
    # - The number of distinct values is significantly less than the total number of examples.
    # - The distinct values are not all numbers.
    # - The distinct values are all numbers but comprise of a perfectly sequential list of integers that suggests the
    #   values represent categories.
    valid_row_count = row_count * (1.0 - missing_value_percent)
    if num_distinct_values < valid_row_count * CATEGORY_TYPE_DISTINCT_VALUE_PERCENTAGE_CUTOFF and (
        (not strings_utils.are_all_numbers(distinct_values)) or strings_utils.are_sequential_integers(distinct_values)
    ):
        return CATEGORY

    # Use NUMBER if all of the distinct values are numbers.
    if strings_utils.are_all_numbers(distinct_values):
        return NUMBER

    # TODO (ASN): add other modalities (image, etc. )
    # Fallback to TEXT.
    return TEXT


@DeveloperAPI
def should_exclude(
    idx: int, field: FieldInfo, dtype: str, column_count: int, row_count: int, targets: Set[str]
) -> bool:
    if field.key == "PRI":
        logging.info(f"Exclude {field.name} ({dtype}): primary key")
        return True

    if field.name in targets:
        return False

    if field.num_distinct_values <= 1:
        logging.info(f"Exclude {field.name} ({dtype}): less than 2 distinct values")
        return True

    distinct_value_percent = float(field.num_distinct_values) / row_count
    if distinct_value_percent == 1.0:
        upper_name = field.name.upper()
        if (
            (idx == 0 and "INDEX" in upper_name and dtype == NUMBER)
            or upper_name.endswith("ID")
            or upper_name.startswith("ID")
        ):
            logging.info(f"Exclude {field.name} ({dtype}): unique ID column")
            return True

    # For TEXT fields, we only want to use them if they appear "interesting", otherwise we would rather exclude
    # them and treat the problem as a tabular problem
    if column_count > 3 and dtype == TEXT and (field.avg_words or 0) < TEXT_AVG_WORDS_CUTOFF:
        logging.info(f"Exclude {field.name} ({dtype}): too few average words")
        return True

    return False



"""Utilities for producing synthetic test data that is convergence-friendly."""

from collections import namedtuple

import numpy as np
import pandas as pd
from sklearn.model_selection import train_test_split

RANDOM_SEED = 42
NUMBER_OBSERVATIONS = 200

GeneratedData = namedtuple("GeneratedData", "train_df validation_df test_df")


def get_feature_configs():
    input_features = [
        {"name": "x", "type": "number"},
    ]
    output_features = [
        {
            "name": "y",
            "type": "number",
            "loss": {"type": "mean_squared_error"},
            "decoder": {
                "num_fc_layers": 2,
                "fc_output_size": 64,
            },
        }
    ]

    return input_features, output_features


def get_generated_data():
    # function generates simple training data that guarantee convergence
    # within 30 epochs for suitable config

    # generate data
    np.random.seed(RANDOM_SEED)
    x = np.array(range(NUMBER_OBSERVATIONS)).reshape(-1, 1)
    y = 2 * x + 1 + np.random.normal(size=x.shape[0]).reshape(-1, 1)
    raw_df = pd.DataFrame(np.concatenate((x, y), axis=1), columns=["x", "y"])

    # create training data
    train, valid_test = train_test_split(raw_df, train_size=0.7)

    # create validation and test data
    validation, test = train_test_split(valid_test, train_size=0.5)

    return GeneratedData(train, validation, test)


def get_generated_data_for_optimizer():
    # function generates simple training data that guarantee convergence
    # within 30 epochs for suitable config

    # generate data
    np.random.seed(RANDOM_SEED)
    x = np.array(range(NUMBER_OBSERVATIONS)).reshape(-1, 1)
    y = 2 * x + 1 + np.random.normal(size=x.shape[0]).reshape(-1, 1)
    raw_df = pd.DataFrame(np.concatenate((x, y), axis=1), columns=["x", "y"])
    raw_df["x"] = (raw_df["x"] - raw_df["x"].min()) / (raw_df["x"].max() - raw_df["x"].min())
    raw_df["y"] = (raw_df["y"] - raw_df["y"].min()) / (raw_df["y"].max() - raw_df["y"].min())

    # create training data
    train, valid_test = train_test_split(raw_df, train_size=0.7)

    # create validation and test data
    validation, test = train_test_split(valid_test, train_size=0.5)

    return GeneratedData(train, validation, test)



import logging
import os
from typing import Optional

from ludwig.constants import CHECKSUM, META, TEST, TRAINING, VALIDATION
from ludwig.data.cache.types import alphanum, CacheableDataset
from ludwig.data.cache.util import calculate_checksum
from ludwig.data.dataset.base import DatasetManager
from ludwig.utils import data_utils
from ludwig.utils.fs_utils import delete, path_exists

logger = logging.getLogger(__name__)


class DatasetCache:
    def __init__(self, config, checksum, cache_map, dataset_manager):
        self.config = config
        self.checksum = checksum
        self.cache_map = cache_map
        self.dataset_manager = dataset_manager

    def get(self):
        training_set_metadata_fp = self.cache_map[META]
        if not path_exists(training_set_metadata_fp):
            return None

        try:
            cached_training_set_metadata = data_utils.load_json(training_set_metadata_fp)
        except Exception:
            logger.exception(f"Failed to load cached training set metadata at {training_set_metadata_fp}")
            return None

        cached_training_set = self.cache_map[TRAINING] if path_exists(self.cache_map[TRAINING]) else None
        if not cached_training_set:
            logger.warning(f"Failed to load cached training set at {self.cache_map[TRAINING]}")

        cached_validation_set = self.cache_map[VALIDATION] if path_exists(self.cache_map[VALIDATION]) else None
        if not cached_validation_set:
            logger.warning(f"Failed to load cached validation set at {self.cache_map[VALIDATION]}")

        cached_test_set = self.cache_map[TEST] if path_exists(self.cache_map[TEST]) else None
        if not cached_test_set:
            logger.warning(f"Failed to load cached test set at {self.cache_map[TEST]}")

        valid = self.checksum == cached_training_set_metadata.get(CHECKSUM) and cached_training_set is not None

        return valid, cached_training_set_metadata, cached_training_set, cached_test_set, cached_validation_set

    def put(self, training_set, test_set, validation_set, training_set_metadata):
        logger.info("Writing preprocessed training set cache")
        training_set = self.dataset_manager.save(
            self.cache_map[TRAINING],
            training_set,
            self.config,
            training_set_metadata,
            TRAINING,
        )

        if test_set is not None:
            logger.info("Writing preprocessed test set cache")
            test_set = self.dataset_manager.save(
                self.cache_map[TEST],
                test_set,
                self.config,
                training_set_metadata,
                TEST,
            )

        if validation_set is not None:
            logger.info("Writing preprocessed validation set cache")
            validation_set = self.dataset_manager.save(
                self.cache_map[VALIDATION],
                validation_set,
                self.config,
                training_set_metadata,
                VALIDATION,
            )

        logger.info("Writing train set metadata")
        data_utils.save_json(self.cache_map[META], training_set_metadata)

        return training_set, test_set, validation_set, training_set_metadata

    def delete(self):
        for fname in self.cache_map.values():
            if path_exists(fname):
                # Parquet entries in the cache_ma can be pointers to directories.
                delete(fname, recursive=True)

    def get_cached_obj_path(self, cached_obj_name: str) -> str:
        return self.cache_map.get(cached_obj_name)


class CacheManager:
    def __init__(
        self,
        dataset_manager: DatasetManager,
        cache_dir: Optional[str] = None,
    ):
        self._dataset_manager = dataset_manager
        self._cache_dir = cache_dir

    def get_dataset_cache(
        self,
        config: dict,
        dataset: Optional[CacheableDataset] = None,
        training_set: Optional[CacheableDataset] = None,
        test_set: Optional[CacheableDataset] = None,
        validation_set: Optional[CacheableDataset] = None,
    ) -> DatasetCache:
        if dataset is not None:
            key = self.get_cache_key(dataset, config)
            cache_map = {
                META: self.get_cache_path(dataset, key, META, "json"),
                TRAINING: self.get_cache_path(dataset, key, TRAINING),
                TEST: self.get_cache_path(dataset, key, TEST),
                VALIDATION: self.get_cache_path(dataset, key, VALIDATION),
            }
            return DatasetCache(config, key, cache_map, self._dataset_manager)
        else:
            key = self.get_cache_key(training_set, config)
            cache_map = {
                META: self.get_cache_path(training_set, key, META, "json"),
                TRAINING: self.get_cache_path(training_set, key, TRAINING),
                TEST: self.get_cache_path(test_set, key, TEST),
                VALIDATION: self.get_cache_path(validation_set, key, VALIDATION),
            }
            return DatasetCache(config, key, cache_map, self._dataset_manager)

    def get_cache_key(self, dataset: CacheableDataset, config: dict) -> str:
        return calculate_checksum(dataset, config)

    def get_cache_path(self, dataset: Optional[CacheableDataset], key: str, tag: str, ext: Optional[str] = None) -> str:
        if self._cache_dir is None and dataset is not None:
            # Use the input dataset filename (minus the extension) as the cache path
            stem = dataset.get_cache_path()
        else:
            # To avoid collisions across different directories, we use the unique checksum
            # as the cache path
            stem = alphanum(key)

        ext = ext or self.data_format
        cache_fname = f"{stem}.{tag}.{ext}"
        return os.path.join(self.get_cache_directory(dataset), cache_fname)

    def get_cache_directory(self, dataset: Optional[CacheableDataset]) -> str:
        if self._cache_dir is None:
            if dataset is None:
                return os.getcwd()
            return dataset.get_cache_directory()
        return self._cache_dir

    def can_cache(self, skip_save_processed_input: bool) -> bool:
        return self._dataset_manager.can_cache(skip_save_processed_input)

    @property
    def data_format(self) -> str:
        return self._dataset_manager.data_format



# from typing import Any, Dict, Optional

# import pytest

# from ludwig.utils.config_utils import merge_fixed_preprocessing_params


# @pytest.mark.parametrize(
#     "pretrained_model_name_or_path",
#     [None, "bert-large-uncased"],
#     ids=["default_model", "override_model"],
# )
# def test_merge_fixed_preprocessing_params(pretrained_model_name_or_path: str):
#     expected_model_name = "bert-base-uncased"

#     preprocessing = {
#         "tokenizer": "space",
#         "lowercase": True,
#     }

#     encoder = {"type": "bert"}
#     if pretrained_model_name_or_path is not None:
#         encoder["pretrained_model_name_or_path"] = pretrained_model_name_or_path
#         expected_model_name = pretrained_model_name_or_path

#     merged_params = merge_fixed_preprocessing_params("text", preprocessing, encoder)
#     assert merged_params == {
#         "tokenizer": "hf_tokenizer",
#         "lowercase": True,
#         "pretrained_model_name_or_path": expected_model_name,
#     }


# @pytest.mark.parametrize(
#     "encoder,expected",
#     [
#         ({"type": "parallel_cnn"}, False),
#         ({"type": "bert", "trainable": False}, None),
#         ({"type": "bert", "trainable": True}, False),
#     ],
#     ids=["parallel_cnn", "bert_fixed", "bert_trainable"],
# )
# def test_merge_fixed_preprocessing_params_cache_embeddings(encoder: Dict[str, Any], expected: Optional[bool]):
#     preprocessing = {
#         "tokenizer": "space",
#         "lowercase": True,
#     }

#     merged_params = merge_fixed_preprocessing_params("text", preprocessing, encoder)
#     assert merged_params.get("cache_encoder_embeddings") == expected



from typing import List, Tuple

import numpy as np
import pytest
import torch
import torchtext

from ludwig.constants import LAST_HIDDEN, LOGITS, SEQUENCE, TEXT, TYPE
from ludwig.features.sequence_feature import _SequencePreprocessing, SequenceInputFeature, SequenceOutputFeature
from ludwig.schema.features.sequence_feature import SequenceInputFeatureConfig, SequenceOutputFeatureConfig
from ludwig.schema.utils import load_config_with_kwargs
from ludwig.utils.torch_utils import get_torch_device
from tests.integration_tests.utils import ENCODERS, sequence_feature

DEVICE = get_torch_device()
BATCH_SIZE = 8
SEQ_SIZE = 6
VOCAB_SIZE = 64


@pytest.fixture(scope="module")
def input_sequence() -> Tuple[torch.Tensor, List]:
    # generates a realistic looking synthetic sequence tensor, i.e.
    # each sequence will have non-zero tokens at the beginning with
    # trailing zero tokens, including a max length token with a single
    # zero token at the end.  Example:
    # [
    #   [3, 5, 6, 0, 0, 0],
    #   [10, 11, 12, 13, 14, 0],   # max length sequence
    #   [32, 0, 0, 0, 0, 0]        # minimum length sequence
    # ]
    input_tensor = torch.zeros([BATCH_SIZE, SEQ_SIZE], dtype=torch.int32).to(DEVICE)
    sequence_lengths = np.random.randint(1, SEQ_SIZE, size=BATCH_SIZE)
    for i in range(input_tensor.shape[0]):
        input_tensor[i, : sequence_lengths[i]] = torch.tensor(
            np.random.randint(2, VOCAB_SIZE, size=sequence_lengths[i])
        )

    # emulate idx2str structure
    idx2str = ["<PAD>", "<UNK>"] + [str(i) for i in range(2, VOCAB_SIZE)]

    return input_tensor, idx2str


@pytest.mark.parametrize("encoder", ENCODERS)
@pytest.mark.parametrize("sequence_type", [SEQUENCE, TEXT])
def test_sequence_input_feature(input_sequence: tuple, encoder: str, sequence_type: str):
    # test assumes "sequence data" has been tokenized and converted to
    # numeric representation.  Focus of this test is primarily on
    # integration with encoder with correctly sized encoder tensor and
    # required properties are present

    input_sequence, idx2str = input_sequence

    # setup input sequence feature definition
    # use sequence_feature() to generate baseline
    # sequence definition and then augment with
    # pre-processing metadata parameters
    input_feature_def = sequence_feature(
        encoder={
            "type": encoder,
            "max_len": SEQ_SIZE,
            # augment with emulated pre-processing metadata
            "max_sequence_length": SEQ_SIZE,
            "vocab": idx2str,
        }
    )
    input_feature_def[TYPE] = sequence_type

    # create sequence input feature object
    sequence_config, _ = load_config_with_kwargs(SequenceInputFeatureConfig, input_feature_def)
    input_feature_obj = SequenceInputFeature(sequence_config).to(DEVICE)

    # confirm dtype property
    assert input_feature_obj.input_dtype == torch.int32

    # confirm input_shape property
    assert input_feature_obj.input_shape == (SEQ_SIZE,)

    # confirm output_shape property default output shape
    # from sequence_feature() function
    encoder_output = input_feature_obj(input_sequence)
    assert encoder_output["encoder_output"].shape == (BATCH_SIZE, *input_feature_obj.output_shape)


@pytest.mark.parametrize("sequence_type", [SEQUENCE, TEXT])
def test_sequence_output_feature(sequence_type: str):
    output_feature_def = sequence_feature(
        decoder={
            "type": "generator",
            "max_len": SEQ_SIZE,
            "max_sequence_length": SEQ_SIZE,
            "vocab_size": VOCAB_SIZE,
        },
        input_size=VOCAB_SIZE,
    )
    output_feature_def[TYPE] = sequence_type
    sequence_config, _ = load_config_with_kwargs(SequenceOutputFeatureConfig, output_feature_def)
    output_feature_obj = SequenceOutputFeature(sequence_config, {}).to(DEVICE)
    combiner_outputs = {
        "combiner_output": torch.randn([BATCH_SIZE, SEQ_SIZE, VOCAB_SIZE], dtype=torch.float32).to(DEVICE)
    }

    text_output = output_feature_obj(combiner_outputs, {})

    assert LAST_HIDDEN in text_output
    assert LOGITS in text_output
    assert text_output[LOGITS].size() == torch.Size([BATCH_SIZE, SEQ_SIZE, VOCAB_SIZE])


def test_sequence_preproc_module_bad_tokenizer():
    metadata = {
        "preprocessing": {
            "lowercase": True,
            "tokenizer": "dutch_lemmatize",
            "unknown_symbol": "<UNK>",
            "padding_symbol": "<PAD>",
            "computed_fill_value": "<UNK>",
        },
        "max_sequence_length": SEQ_SIZE,
        "str2idx": {"<EOS>": 0, "<SOS>": 1, "<PAD>": 2, "<UNK>": 3, "▁hell": 4, "o": 5, "▁world": 6},
    }

    with pytest.raises(ValueError):
        _SequencePreprocessing(metadata)


def test_sequence_preproc_module_space_tokenizer():
    metadata = {
        "preprocessing": {
            "lowercase": True,
            "tokenizer": "space",
            "unknown_symbol": "<UNK>",
            "padding_symbol": "<PAD>",
            "computed_fill_value": "<UNK>",
        },
        "max_sequence_length": SEQ_SIZE,
        "str2idx": {
            "<EOS>": 0,
            "<SOS>": 1,
            "<PAD>": 2,
            "<UNK>": 3,
            "hello": 4,
            "world": 5,
            "paleontology": 6,
        },
    }
    module = _SequencePreprocessing(metadata)

    res = module(["    paleontology", "unknown", "hello    world hello", "hello world hello     world    "])

    assert torch.allclose(
        res, torch.tensor([[1, 6, 0, 2, 2, 2], [1, 3, 0, 2, 2, 2], [1, 4, 5, 4, 0, 2], [1, 4, 5, 4, 5, 0]])
    )


def test_text_preproc_module_space_punct_tokenizer():
    metadata = {
        "preprocessing": {
            "lowercase": True,
            "tokenizer": "space_punct",
            "unknown_symbol": "<UNK>",
            "padding_symbol": "<PAD>",
            "computed_fill_value": "<UNK>",
        },
        "max_sequence_length": SEQ_SIZE,
        "str2idx": {
            "<EOS>": 0,
            "<SOS>": 1,
            "<PAD>": 2,
            "<UNK>": 3,
            "this": 4,
            "sentence": 5,
            "has": 6,
            "punctuation": 7,
            ",": 8,
            ".": 9,
        },
    }
    module = _SequencePreprocessing(metadata)

    res = module(["punctuation", ",,,,", "this... this... punctuation", "unknown"])

    assert torch.allclose(
        res, torch.tensor([[1, 7, 0, 2, 2, 2], [1, 8, 8, 8, 8, 0], [1, 4, 9, 9, 9, 4], [1, 3, 0, 2, 2, 2]])
    )


@pytest.mark.skipif(
    torch.torch_version.TorchVersion(torchtext.__version__) < (0, 12, 0), reason="requires torchtext 0.12.0 or higher"
)
def test_sequence_preproc_module_sentencepiece_tokenizer():
    metadata = {
        "preprocessing": {
            "lowercase": True,
            "tokenizer": "sentencepiece",
            "unknown_symbol": "<UNK>",
            "padding_symbol": "<PAD>",
            "computed_fill_value": "<UNK>",
        },
        "max_sequence_length": SEQ_SIZE,
        "str2idx": {
            "<EOS>": 0,
            "<SOS>": 1,
            "<PAD>": 2,
            "<UNK>": 3,
            "▁hell": 4,
            "o": 5,
            "▁world": 6,
            "▁pale": 7,
            "ont": 8,
            "ology": 9,
        },
    }
    module = _SequencePreprocessing(metadata)

    res = module(["paleontology", "unknown", "hello world hello", "hello world hello world"])

    assert torch.allclose(
        res, torch.tensor([[1, 7, 8, 9, 0, 2], [1, 3, 3, 3, 0, 2], [1, 4, 5, 6, 4, 5], [1, 4, 5, 6, 4, 5]])
    )


@pytest.mark.skipif(
    torch.torch_version.TorchVersion(torchtext.__version__) < (0, 12, 0), reason="requires torchtext 0.12.0 or higher"
)
def test_sequence_preproc_module_clip_tokenizer():
    metadata = {
        "preprocessing": {
            "lowercase": True,
            "tokenizer": "clip",
            "unknown_symbol": "<UNK>",
            "padding_symbol": "<PAD>",
            "computed_fill_value": "<UNK>",
        },
        "max_sequence_length": SEQ_SIZE,
        "str2idx": {
            "<EOS>": 0,
            "<SOS>": 1,
            "<PAD>": 2,
            "<UNK>": 3,
            "hello</w>": 4,
            "world</w>": 5,
            "pale": 7,
            "ontology</w>": 8,
        },
    }
    module = _SequencePreprocessing(metadata)

    res = module(["paleontology", "unknown", "hello world hello", "hello world hello world"])

    assert torch.allclose(
        res, torch.tensor([[1, 7, 8, 0, 2, 2], [1, 3, 0, 2, 2, 2], [1, 4, 5, 4, 0, 2], [1, 4, 5, 4, 5, 0]])
    )


@pytest.mark.skipif(
    torch.torch_version.TorchVersion(torchtext.__version__) < (0, 12, 0), reason="requires torchtext 0.12.0 or higher"
)
def test_sequence_preproc_module_gpt2bpe_tokenizer():
    metadata = {
        "preprocessing": {
            "lowercase": True,
            "tokenizer": "gpt2bpe",
            "unknown_symbol": "<UNK>",
            "padding_symbol": "<PAD>",
            "computed_fill_value": "<UNK>",
        },
        "max_sequence_length": SEQ_SIZE,
        "str2idx": {
            "<EOS>": 0,
            "<SOS>": 1,
            "<PAD>": 2,
            "<UNK>": 3,
            "hello": 4,
            "Ġworld": 5,
            "Ġhello": 7,
            "p": 8,
            "ale": 9,
            "ont": 10,
            "ology": 11,
        },
    }
    module = _SequencePreprocessing(metadata)

    res = module(["paleontology", "unknown", "hello world hello", "hello world hello world"])

    assert torch.allclose(
        res, torch.tensor([[1, 8, 9, 10, 11, 0], [1, 3, 0, 2, 2, 2], [1, 4, 5, 7, 0, 2], [1, 4, 5, 7, 5, 0]])
    )


@pytest.mark.skipif(
    torch.torch_version.TorchVersion(torchtext.__version__) < (0, 13, 0), reason="requires torchtext 0.13.0 or higher"
)
def test_sequence_preproc_module_bert_tokenizer():
    metadata = {
        "preprocessing": {
            "lowercase": True,
            "tokenizer": "bert",
            "unknown_symbol": "<UNK>",
            "padding_symbol": "<PAD>",
            "computed_fill_value": "<UNK>",
        },
        "max_sequence_length": SEQ_SIZE,
        "str2idx": {
            "<EOS>": 0,
            "<SOS>": 1,
            "<PAD>": 2,
            "<UNK>": 3,
            "hello": 4,
            "world": 5,
            "pale": 7,
            "##ont": 8,
            "##ology": 9,
        },
    }
    module = _SequencePreprocessing(metadata)

    res = module(["paleontology", "unknown", "hello world hello", "hello world hello world"])

    assert torch.allclose(
        res, torch.tensor([[1, 7, 8, 9, 0, 2], [1, 3, 0, 2, 2, 2], [1, 4, 5, 4, 0, 2], [1, 4, 5, 4, 5, 0]])
    )



#! /usr/bin/env python
# Copyright (c) 2019 Uber Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import logging
from abc import ABC
from functools import lru_cache
from typing import Dict

import torch
from torch.nn import Linear, ModuleList

from ludwig.api_annotations import DeveloperAPI
from ludwig.constants import BINARY, NUMBER
from ludwig.encoders.registry import get_sequence_encoder_registry
from ludwig.features.base_feature import InputFeature
from ludwig.modules.attention_modules import TransformerStack
from ludwig.modules.embedding_modules import Embed
from ludwig.modules.fully_connected_modules import FCStack
from ludwig.modules.reduction_modules import SequenceReducer
from ludwig.modules.tabnet_modules import TabNet
from ludwig.schema.combiners.comparator import ComparatorCombinerConfig
from ludwig.schema.combiners.concat import ConcatCombinerConfig
from ludwig.schema.combiners.project_aggregate import ProjectAggregateCombinerConfig
from ludwig.schema.combiners.sequence import SequenceCombinerConfig
from ludwig.schema.combiners.sequence_concat import SequenceConcatCombinerConfig
from ludwig.schema.combiners.tab_transformer import TabTransformerCombinerConfig
from ludwig.schema.combiners.tabnet import TabNetCombinerConfig
from ludwig.schema.combiners.transformer import TransformerCombinerConfig
from ludwig.schema.combiners.utils import combiner_registry, register_combiner
from ludwig.utils.misc_utils import get_from_registry
from ludwig.utils.torch_utils import LudwigModule, sequence_length_3D
from ludwig.utils.torch_utils import sequence_mask as torch_sequence_mask

logger = logging.getLogger(__name__)


def get_combiner_class(combiner_type: str):
    """Returns the corresponding combiner class from `ludwig.combiners.combiners.combiner_registry`.

    :param combiner_type: identifier that should correspond to a registered combiner
    :type combiner_type: str
    """
    return get_from_registry(combiner_type, combiner_registry)


@DeveloperAPI
class Combiner(LudwigModule, ABC):
    """Base class for combiners, which implements common properties.

    Subclasses will usually override:     __init__()        to set properties and allocate resources.  Should call
    super().__init__(input_features).     forward()         performs the forward pass given a dictionary of encoder
    outputs.     get_schema_cls()  must returns the class of the corresponding schema for the combiner type.
    """

    def __init__(self, input_features: Dict[str, "InputFeature"]):
        super().__init__()
        self.input_features = input_features

    @property
    def concatenated_shape(self) -> torch.Size:
        # compute the size of the last dimension for the incoming encoder outputs
        # this is required to setup the fully connected layer
        shapes = [torch.prod(torch.Tensor([*self.input_features.get(k).output_shape])) for k in self.input_features]
        return torch.Size([torch.sum(torch.Tensor(shapes)).type(torch.int32)])

    @property
    def input_shape(self) -> Dict:
        # input to combiner is a dictionary of the input features encoder
        # outputs, this property returns dictionary of output shapes for each
        # input feature's encoder output shapes.
        return {k: self.input_features.get(k).output_shape for k in self.input_features}

    @property
    @lru_cache(maxsize=1)
    def output_shape(self) -> torch.Size:
        pseudo_input = {}
        for k in self.input_features:
            pseudo_input[k] = {
                "encoder_output": torch.rand(
                    2, *self.input_features.get(k).output_shape, dtype=self.input_dtype, device=self.device
                )
            }
        output_tensor = self.forward(pseudo_input)
        return output_tensor["combiner_output"].size()[1:]


@register_combiner(name="concat")
class ConcatCombiner(Combiner):
    def __init__(self, input_features: Dict[str, "InputFeature"] = None, config: ConcatCombinerConfig = None, **kwargs):
        super().__init__(input_features)
        self.name = "ConcatCombiner"
        logger.debug(f" {self.name}")

        self.flatten_inputs = config.flatten_inputs
        self.fc_stack = None

        # todo future: this may be redundant, check
        fc_layers = config.fc_layers
        if fc_layers is None:
            fc_layers = []
            for i in range(config.num_fc_layers):
                fc_layers.append({"output_size": config.output_size})
        self.fc_layers = fc_layers

        logger.debug("  FCStack")
        self.fc_stack = FCStack(
            first_layer_input_size=self.concatenated_shape[-1],
            layers=config.fc_layers,
            num_layers=config.num_fc_layers,
            default_output_size=config.output_size,
            default_use_bias=config.use_bias,
            default_weights_initializer=config.weights_initializer,
            default_bias_initializer=config.bias_initializer,
            default_norm=config.norm,
            default_norm_params=config.norm_params,
            default_activation=config.activation,
            default_dropout=config.dropout,
            residual=config.residual,
        )

        if input_features and len(input_features) == 1 and self.fc_layers is None:
            self.supports_masking = True

    def forward(self, inputs: Dict) -> Dict:  # encoder outputs
        encoder_outputs = [inputs[k]["encoder_output"] for k in inputs]

        # ================ Flatten ================
        if self.flatten_inputs:
            batch_size = encoder_outputs[0].shape[0]
            encoder_outputs = [torch.reshape(eo, [batch_size, -1]) for eo in encoder_outputs]

        # ================ Concat ================
        if len(encoder_outputs) > 1:
            hidden = torch.cat(encoder_outputs, 1)
        else:
            hidden = list(encoder_outputs)[0]

        # ================ Fully Connected ================
        hidden = self.fc_stack(hidden)

        return_data = {"combiner_output": hidden}

        if len(inputs) == 1:
            # Workaround for including additional tensors from output of input encoders for
            # potential use in decoders, e.g. LSTM state for seq2seq.
            # TODO(Justin): Think about how to make this communication work for multi-sequence
            # features. Other combiners.
            for key, value in [d for d in inputs.values()][0].items():
                if key != "encoder_output":
                    return_data[key] = value

        return return_data

    @staticmethod
    def get_schema_cls():
        return ConcatCombinerConfig


@register_combiner(name="sequence_concat")
class SequenceConcatCombiner(Combiner):
    def __init__(
        self, input_features: Dict[str, "InputFeature"], config: SequenceConcatCombinerConfig = None, **kwargs
    ):
        super().__init__(input_features)
        self.name = "SequenceConcatCombiner"
        logger.debug(f" {self.name}")

        self.reduce_output = config.reduce_output
        self.reduce_sequence = SequenceReducer(
            reduce_mode=config.reduce_output,
            max_sequence_length=self.concatenated_shape[0],
            encoding_size=self.concatenated_shape[1],
        )
        if self.reduce_output is None:
            self.supports_masking = True
        self.main_sequence_feature = config.main_sequence_feature

    @property
    def concatenated_shape(self) -> torch.Size:
        # computes the effective shape of the input tensor after combining
        # all the encoder outputs
        # determine max sequence length by finding the first sequence tensor
        # assume all the sequences are of the same size, if not true
        # this will be caught during processing
        seq_size = None
        for k in self.input_features:
            # dim-2 output_shape implies a sequence [seq_size, hidden]
            if len(self.input_features.get(k).output_shape) == 2:
                seq_size = self.input_features.get(k).output_shape[0]
                break

        # collect the size of the last dimension for all input feature
        # encoder outputs
        shapes = [
            self.input_features.get(k).output_shape[-1] for k in self.input_features
        ]  # output shape not input shape
        return torch.Size([seq_size, sum(shapes)])

    def forward(self, inputs: Dict) -> Dict:  # encoder outputs
        if self.main_sequence_feature is None or self.main_sequence_feature not in inputs:
            for if_name, if_outputs in inputs.items():
                # todo: when https://github.com/ludwig-ai/ludwig/issues/810 is closed
                #       convert following test from using shape to use explicit
                #       if_outputs[TYPE] values for sequence features
                if len(if_outputs["encoder_output"].shape) == 3:
                    self.main_sequence_feature = if_name
                    break

        if self.main_sequence_feature is None:
            raise Exception("No sequence feature available for sequence combiner")

        main_sequence_feature_encoding = inputs[self.main_sequence_feature]

        representation = main_sequence_feature_encoding["encoder_output"]
        representations = [representation]

        sequence_max_length = representation.shape[1]
        sequence_length = sequence_length_3D(representation)

        # ================ Concat ================
        for if_name, if_outputs in inputs.items():
            if if_name != self.main_sequence_feature:
                if_representation = if_outputs["encoder_output"]
                if len(if_representation.shape) == 3:
                    # The following check makes sense when
                    # both representations have a specified
                    # sequence length dimension. If they do not,
                    # then this check is simply checking if None == None
                    # and will not catch discrepancies in the different
                    # feature length dimension. Those errors will show up
                    # at training time. Possible solutions to this is
                    # to enforce a length second dimension in
                    # sequential feature placeholders, but that
                    # does not work with BucketedBatcher that requires
                    # the second dimension to be undefined in order to be
                    # able to trim the data points and speed up computation.
                    # So for now we are keeping things like this, make sure
                    # to write in the documentation that training time
                    # dimensions mismatch may occur if the sequential
                    # features have different lengths for some data points.
                    if if_representation.shape[1] != representation.shape[1]:
                        raise ValueError(
                            "The sequence length of the input feature {} "
                            "is {} and is different from the sequence "
                            "length of the main sequence feature {} which "
                            "is {}.\n Shape of {}: {}, shape of {}: {}.\n"
                            "Sequence lengths of all sequential features "
                            "must be the same  in order to be concatenated "
                            "by the sequence concat combiner. "
                            "Try to impose the same max sequence length "
                            "as a preprocessing parameter to both features "
                            "or to reduce the output of {}.".format(
                                if_name,
                                if_representation.shape[1],
                                self.main_sequence_feature,
                                representation.shape[1],
                                if_name,
                                if_representation.shape,
                                if_name,
                                representation.shape,
                                if_name,
                            )
                        )
                    # this assumes all sequence representations have the
                    # same sequence length, 2nd dimension
                    representations.append(if_representation)

                elif len(if_representation.shape) == 2:
                    multipliers = (1, sequence_max_length, 1)
                    tiled_representation = torch.tile(torch.unsqueeze(if_representation, 1), multipliers)
                    representations.append(tiled_representation)

                else:
                    raise ValueError(
                        "The representation of {} has rank {} and cannot be"
                        " concatenated by a sequence concat combiner. "
                        "Only rank 2 and rank 3 tensors are supported.".format(if_name, len(if_representation.shape))
                    )

        hidden = torch.cat(representations, 2)
        logger.debug(f"  concat_hidden: {hidden}")

        # ================ Mask ================
        sequence_mask = torch_sequence_mask(sequence_length, sequence_max_length)
        hidden = torch.multiply(hidden, torch.unsqueeze(sequence_mask, -1).type(torch.float32))

        # ================ Reduce ================
        hidden = self.reduce_sequence(hidden)

        return_data = {"combiner_output": hidden}

        if len(inputs) == 1:
            for key, value in [d for d in inputs.values()][0].items():
                if key != "encoder_output":
                    return_data[key] = value

        return return_data

    @staticmethod
    def get_schema_cls():
        return SequenceConcatCombinerConfig


@register_combiner(name="sequence")
class SequenceCombiner(Combiner):
    def __init__(self, input_features: Dict[str, "InputFeature"], config: SequenceCombinerConfig = None, **kwargs):
        super().__init__(input_features)
        self.name = "SequenceCombiner"
        logger.debug(f" {self.name}")

        self.combiner = SequenceConcatCombiner(
            input_features,
            config=SequenceConcatCombinerConfig(reduce_output=None, main_sequence_feature=config.main_sequence_feature),
        )

        logger.debug(
            f"combiner input shape {self.combiner.concatenated_shape}, " f"output shape {self.combiner.output_shape}"
        )

        self.encoder_obj = get_from_registry(config.encoder.type, get_sequence_encoder_registry())(
            should_embed=False,
            reduce_output=config.reduce_output,
            embedding_size=self.combiner.output_shape[1],
            max_sequence_length=self.combiner.output_shape[0],
            **kwargs,
        )

        if hasattr(self.encoder_obj, "supports_masking") and self.encoder_obj.supports_masking:
            self.supports_masking = True

    @property
    def concatenated_shape(self) -> torch.Size:
        # computes the effective shape of the input tensor after combining
        # all the encoder outputs
        # determine max sequence length by finding the first sequence tensor
        # assume all the sequences are of the same size, if not true
        # this will be caught during processing
        seq_size = None
        for k in self.input_features:
            # dim-2 output_shape implies a sequence [seq_size, hidden]
            if len(self.input_features.get(k).output_shape) == 2:
                seq_size = self.input_features.get(k).output_shape[0]
                break

        # collect the size of the last dimension for all input feature
        # encoder outputs
        shapes = [
            self.input_features.get(k).output_shape[-1] for k in self.input_features
        ]  # output shape not input shape
        return torch.Size([seq_size, sum(shapes)])

    def forward(self, inputs: Dict) -> Dict:  # encoder outputs
        # ================ Concat ================
        hidden = self.combiner(inputs)

        # ================ Sequence encoding ================
        hidden = self.encoder_obj(hidden["combiner_output"])

        return_data = {"combiner_output": hidden["encoder_output"]}
        for key, value in hidden.items():
            if key != "encoder_output":
                return_data[key] = value

        return return_data

    @staticmethod
    def get_schema_cls():
        return SequenceCombinerConfig


@register_combiner(name="tabnet")
class TabNetCombiner(Combiner):
    def __init__(
        self, input_features: Dict[str, "InputFeature"], config: TabNetCombinerConfig = None, **kwargs
    ) -> None:
        super().__init__(input_features)
        self.name = "TabNetCombiner"
        logger.debug(f" {self.name}")

        self.tabnet = TabNet(
            self.concatenated_shape[-1],
            config.size,
            config.output_size,
            num_steps=config.num_steps,
            num_total_blocks=config.num_total_blocks,
            num_shared_blocks=config.num_shared_blocks,
            relaxation_factor=config.relaxation_factor,
            bn_epsilon=config.bn_epsilon,
            bn_momentum=config.bn_momentum,
            bn_virtual_bs=config.bn_virtual_bs,
            sparsity=config.sparsity,
            entmax_mode=config.entmax_mode,
            entmax_alpha=config.entmax_alpha,
        )

        if config.dropout > 0:
            self.dropout = torch.nn.Dropout(config.dropout)
        else:
            self.dropout = None

    @property
    def concatenated_shape(self) -> torch.Size:
        # compute the size of the last dimension for the incoming encoder outputs
        # this is required to setup
        shapes = [torch.prod(torch.Tensor([*self.input_features.get(k).output_shape])) for k in self.input_features]
        return torch.Size([torch.sum(torch.Tensor(shapes)).type(torch.int32)])

    def forward(
        self,
        inputs: torch.Tensor,  # encoder outputs
    ) -> Dict:
        encoder_outputs = [inputs[k]["encoder_output"] for k in inputs]

        # ================ Flatten ================
        batch_size = encoder_outputs[0].shape[0]
        encoder_outputs = [torch.reshape(eo, [batch_size, -1]) for eo in encoder_outputs]

        # ================ Concat ================
        if len(encoder_outputs) > 1:
            hidden = torch.cat(encoder_outputs, 1)
        else:
            hidden = list(encoder_outputs)[0]

        # ================ TabNet ================
        hidden, aggregated_mask, masks = self.tabnet(hidden)
        if self.dropout:
            hidden = self.dropout(hidden)

        return_data = {
            "combiner_output": hidden,
            "aggregated_attention_masks": aggregated_mask,
            "attention_masks": masks,
        }

        if len(inputs) == 1:
            for key, value in [d for d in inputs.values()][0].items():
                if key != "encoder_output":
                    return_data[key] = value

        return return_data

    @staticmethod
    def get_schema_cls():
        return TabNetCombinerConfig

    @property
    def output_shape(self) -> torch.Size:
        return self.tabnet.output_shape


@register_combiner(name="transformer")
class TransformerCombiner(Combiner):
    def __init__(
        self, input_features: Dict[str, "InputFeature"] = None, config: TransformerCombinerConfig = None, **kwargs
    ):
        super().__init__(input_features)
        self.name = "TransformerCombiner"
        logger.debug(f" {self.name}")

        self.reduce_output = config.reduce_output
        self.reduce_sequence = SequenceReducer(
            reduce_mode=config.reduce_output,
            max_sequence_length=len(self.input_features),
            encoding_size=config.hidden_size,
        )
        if self.reduce_output is None:
            self.supports_masking = True

        # max sequence length for Transformer layer is number of input features
        self.max_sequence_length = len(self.input_features)

        logger.debug("  Projectors")
        self.projectors = ModuleList(
            # regardless of rank-2 or rank-3 input, torch.prod() calculates size
            # after flattening the encoder output tensor
            [
                Linear(
                    torch.prod(torch.Tensor([*input_features[inp].output_shape])).type(torch.int32), config.hidden_size
                )
                for inp in input_features
            ]
        )

        logger.debug("  TransformerStack")
        self.transformer_stack = TransformerStack(
            input_size=config.hidden_size,
            max_sequence_length=self.max_sequence_length,
            hidden_size=config.hidden_size,
            num_heads=config.num_heads,
            output_size=config.transformer_output_size,
            num_layers=config.num_layers,
            dropout=config.dropout,
        )

        if self.reduce_output is not None:
            logger.debug("  FCStack")
            self.fc_stack = FCStack(
                self.transformer_stack.output_shape[-1],
                layers=config.fc_layers,
                num_layers=config.num_fc_layers,
                default_output_size=config.output_size,
                default_use_bias=config.use_bias,
                default_weights_initializer=config.weights_initializer,
                default_bias_initializer=config.bias_initializer,
                default_norm=config.norm,
                default_norm_params=config.norm_params,
                default_activation=config.fc_activation,
                default_dropout=config.fc_dropout,
                fc_residual=config.fc_residual,
            )

    def forward(
        self,
        inputs,  # encoder outputs
    ) -> Dict:
        encoder_outputs = [inputs[k]["encoder_output"] for k in inputs]

        # ================ Flatten ================
        batch_size = encoder_outputs[0].shape[0]
        encoder_outputs = [torch.reshape(eo, [batch_size, -1]) for eo in encoder_outputs]

        # ================ Project & Concat ================
        projected = [self.projectors[i](eo) for i, eo in enumerate(encoder_outputs)]
        hidden = torch.stack(projected)  # shape [num_eo, bs, h]
        hidden = torch.permute(hidden, (1, 0, 2))  # shape [bs, num_eo, h]

        # ================ Transformer Layers ================
        hidden = self.transformer_stack(hidden)

        # ================ Sequence Reduction ================
        if self.reduce_output is not None:
            hidden = self.reduce_sequence(hidden)

            # ================ FC Layers ================
            hidden = self.fc_stack(hidden)

        return_data = {"combiner_output": hidden}

        if len(inputs) == 1:
            for key, value in [d for d in inputs.values()][0].items():
                if key != "encoder_output":
                    return_data[key] = value

        return return_data

    @staticmethod
    def get_schema_cls():
        return TransformerCombinerConfig


@register_combiner(name="tabtransformer")
class TabTransformerCombiner(Combiner):
    def __init__(
        self, input_features: Dict[str, "InputFeature"] = None, config: TabTransformerCombinerConfig = None, **kwargs
    ):
        super().__init__(input_features)
        self.name = "TabTransformerCombiner"
        logger.debug(f"Initializing {self.name}")

        self.reduce_output = config.reduce_output
        self.reduce_sequence = SequenceReducer(
            reduce_mode=config.reduce_output, max_sequence_length=len(input_features), encoding_size=config.hidden_size
        )
        self.supports_masking = True

        self.embed_input_feature_name = config.embed_input_feature_name
        if self.embed_input_feature_name:
            vocab = [
                i_f
                for i_f in input_features
                if input_features[i_f].type() != NUMBER or input_features[i_f].type() != BINARY
            ]
            if self.embed_input_feature_name == "add":
                self.embed_i_f_name_layer = Embed(vocab, config.hidden_size, force_embedding_size=True)
                projector_size = config.hidden_size
            elif isinstance(self.embed_input_feature_name, int):
                if self.embed_input_feature_name > config.hidden_size:
                    raise ValueError(
                        "TabTransformer parameter "
                        "`embed_input_feature_name` "
                        "specified integer value ({}) "
                        "needs to be smaller than "
                        "`hidden_size` ({}).".format(self.embed_input_feature_name, config.hidden_size)
                    )
                self.embed_i_f_name_layer = Embed(
                    vocab,
                    self.embed_input_feature_name,
                    force_embedding_size=True,
                )
                projector_size = config.hidden_size - self.embed_input_feature_name
            else:
                raise ValueError(
                    "TabTransformer parameter "
                    "`embed_input_feature_name` "
                    "should be either None, an integer or `add`, "
                    "the current value is "
                    "{}".format(self.embed_input_feature_name)
                )
        else:
            projector_size = config.hidden_size

        logger.debug("  Projectors")
        self.unembeddable_features = []
        self.embeddable_features = []
        for i_f in input_features:
            if input_features[i_f].type() in {NUMBER, BINARY}:
                self.unembeddable_features.append(i_f)
            else:
                self.embeddable_features.append(i_f)

        self.projectors = ModuleList()
        for i_f in self.embeddable_features:
            flatten_size = self.get_flatten_size(input_features[i_f].output_shape)
            self.projectors.append(Linear(flatten_size[0], projector_size))

        # input to layer_norm are the encoder outputs for unembeddable features,
        # which are number or binary features.  These should be 2-dim
        # tensors.  Size should be concatenation of these tensors.
        concatenated_unembeddable_encoders_size = 0
        for i_f in self.unembeddable_features:
            concatenated_unembeddable_encoders_size += input_features[i_f].output_shape[0]

        self.layer_norm = torch.nn.LayerNorm(concatenated_unembeddable_encoders_size)

        logger.debug("  TransformerStack")
        self.transformer_stack = TransformerStack(
            input_size=config.hidden_size,
            max_sequence_length=len(self.embeddable_features),
            hidden_size=config.hidden_size,
            # todo: can we just use projector_size? # hidden_size,
            num_heads=config.num_heads,
            output_size=config.transformer_output_size,
            num_layers=config.num_layers,
            dropout=config.dropout,
        )

        logger.debug("  FCStack")

        # determine input size to fully connected layer based on reducer
        if config.reduce_output == "concat":
            fc_input_size = len(self.embeddable_features) * config.hidden_size
        else:
            fc_input_size = self.reduce_sequence.output_shape[-1] if len(self.embeddable_features) > 0 else 0
        self.fc_stack = FCStack(
            fc_input_size + concatenated_unembeddable_encoders_size,
            layers=config.fc_layers,
            num_layers=config.num_fc_layers,
            default_output_size=config.output_size,
            default_use_bias=config.use_bias,
            default_weights_initializer=config.weights_initializer,
            default_bias_initializer=config.bias_initializer,
            default_norm=config.norm,
            default_norm_params=config.norm_params,
            default_activation=config.fc_activation,
            default_dropout=config.fc_dropout,
            fc_residual=config.fc_residual,
        )

        self._empty_hidden = torch.empty([1, 0])
        self._embeddable_features_indices = torch.arange(0, len(self.embeddable_features))

        # Create empty tensor of shape [1, 0] to use as hidden in case there are no category or numeric/binary features.
        self.register_buffer("empty_hidden", self._empty_hidden)
        self.register_buffer("embeddable_features_indices", self._embeddable_features_indices)

    @staticmethod
    def get_flatten_size(output_shape: torch.Size) -> torch.Size:
        size = torch.prod(torch.Tensor([*output_shape]))
        return torch.Size([size.type(torch.int32)])

    @property
    def output_shape(self) -> torch.Size:
        return self.fc_stack.output_shape

    def forward(
        self,
        inputs: Dict,  # encoder outputs
    ) -> Dict:
        unembeddable_encoder_outputs = [inputs[k]["encoder_output"] for k in inputs if k in self.unembeddable_features]
        embeddable_encoder_outputs = [inputs[k]["encoder_output"] for k in inputs if k in self.embeddable_features]

        batch_size = (
            embeddable_encoder_outputs[0].shape[0]
            if len(embeddable_encoder_outputs) > 0
            else unembeddable_encoder_outputs[0].shape[0]
        )

        # ================ Project & Concat embeddables ================
        if len(embeddable_encoder_outputs) > 0:

            # ============== Flatten =================
            embeddable_encoder_outputs = [torch.reshape(eo, [batch_size, -1]) for eo in embeddable_encoder_outputs]

            projected = [self.projectors[i](eo) for i, eo in enumerate(embeddable_encoder_outputs)]
            hidden = torch.stack(projected)  # num_eo, bs, h
            hidden = torch.permute(hidden, (1, 0, 2))  # bs, num_eo, h

            if self.embed_input_feature_name:
                i_f_names_idcs = torch.reshape(
                    torch.arange(0, len(embeddable_encoder_outputs), device=self.device), [-1, 1]
                )
                embedded_i_f_names = self.embed_i_f_name_layer(i_f_names_idcs)
                embedded_i_f_names = torch.unsqueeze(embedded_i_f_names, dim=0)
                embedded_i_f_names = torch.tile(embedded_i_f_names, [batch_size, 1, 1])
                if self.embed_input_feature_name == "add":
                    hidden = hidden + embedded_i_f_names
                else:
                    hidden = torch.cat([hidden, embedded_i_f_names], -1)

            # ================ Transformer Layers ================
            hidden = self.transformer_stack(hidden)

            # ================ Sequence Reduction ================
            hidden = self.reduce_sequence(hidden)
        else:
            # create empty tensor because there are no category features
            hidden = torch.empty([batch_size, 0], device=self.device)

        # ================ Concat Skipped ================
        if len(unembeddable_encoder_outputs) > 0:
            unembeddable_encoder_outputs = [torch.reshape(eo, [batch_size, -1]) for eo in unembeddable_encoder_outputs]
            # ================ Flatten ================
            if len(unembeddable_encoder_outputs) > 1:
                unembeddable_hidden = torch.cat(unembeddable_encoder_outputs, -1)  # tf.keras.layers.concatenate
            else:
                unembeddable_hidden = list(unembeddable_encoder_outputs)[0]
            unembeddable_hidden = self.layer_norm(unembeddable_hidden)

        else:
            # create empty tensor because there are not numeric/binary features
            unembeddable_hidden = torch.tile(self.empty_hidden, [batch_size, 0])

        # ================ Concat Skipped and Others ================
        hidden = torch.cat([hidden, unembeddable_hidden], -1)

        # ================ FC Layers ================
        hidden = self.fc_stack(hidden)

        return_data = {"combiner_output": hidden}

        if len(inputs) == 1:
            for key, value in [d for d in inputs.values()][0].items():
                if key != "encoder_output":
                    return_data[key] = value

        return return_data

    @staticmethod
    def get_schema_cls():
        return TabTransformerCombinerConfig


@register_combiner(name="comparator")
class ComparatorCombiner(Combiner):
    def __init__(
        self,
        input_features: Dict[str, "InputFeature"],
        config: ComparatorCombinerConfig = None,
        **kwargs,
    ):
        super().__init__(input_features)
        self.name = "ComparatorCombiner"
        logger.debug(f"Entering {self.name}")

        self.entity_1 = config.entity_1
        self.entity_2 = config.entity_2
        self.required_inputs = set(config.entity_1 + config.entity_2)
        self.output_size = config.output_size

        self.fc_stack = None

        # todo future: this may be redundant, check
        fc_layers = config.fc_layers
        if fc_layers is None and config.num_fc_layers is not None:
            fc_layers = []
            for _ in range(config.num_fc_layers):
                fc_layers.append({"output_size": config.output_size})

        if fc_layers is not None:
            logger.debug("Setting up FCStack")
            self.e1_fc_stack = FCStack(
                self.get_entity_shape(config.entity_1)[-1],
                layers=fc_layers,
                num_layers=config.num_fc_layers,
                default_output_size=config.output_size,
                default_use_bias=config.use_bias,
                default_weights_initializer=config.weights_initializer,
                default_bias_initializer=config.bias_initializer,
                default_norm=config.norm,
                default_norm_params=config.norm_params,
                default_activation=config.activation,
                default_dropout=config.dropout,
            )
            self.e2_fc_stack = FCStack(
                self.get_entity_shape(config.entity_2)[-1],
                layers=fc_layers,
                num_layers=config.num_fc_layers,
                default_output_size=config.output_size,
                default_use_bias=config.use_bias,
                default_weights_initializer=config.weights_initializer,
                default_bias_initializer=config.bias_initializer,
                default_norm=config.norm,
                default_norm_params=config.norm_params,
                default_activation=config.activation,
                default_dropout=config.dropout,
            )

        self.last_fc_layer_output_size = fc_layers[-1]["output_size"]

        # todo: set initializer and regularization
        self.register_buffer(
            "bilinear_weights",
            torch.randn([self.last_fc_layer_output_size, self.last_fc_layer_output_size], dtype=torch.float32),
        )

    def get_entity_shape(self, entity: list) -> torch.Size:
        sizes = [torch.prod(torch.Tensor([*self.input_features.get(k).output_shape])) for k in entity]
        return torch.Size([torch.sum(torch.Tensor(sizes)).type(torch.int32)])

    @property
    def output_shape(self) -> torch.Size:
        return torch.Size([2 * self.last_fc_layer_output_size + 2])

    def forward(
        self,
        inputs: Dict,  # encoder outputs
    ) -> Dict[str, torch.Tensor]:  # encoder outputs
        if inputs.keys() != self.required_inputs:
            raise ValueError(f"Missing inputs {self.required_inputs - set(inputs.keys())}")

        ############
        # Entity 1 #
        ############
        e1_enc_outputs = [inputs[k]["encoder_output"] for k in self.entity_1]

        # ================ Flatten ================
        batch_size = e1_enc_outputs[0].shape[0]
        e1_enc_outputs = [torch.reshape(eo, [batch_size, -1]) for eo in e1_enc_outputs]

        # ================ Concat ================
        if len(e1_enc_outputs) > 1:
            e1_hidden = torch.cat(e1_enc_outputs, 1)
        else:
            e1_hidden = list(e1_enc_outputs)[0]

        # ================ Fully Connected ================
        e1_hidden = self.e1_fc_stack(e1_hidden)  # [bs, output_size]

        ############
        # Entity 2 #
        ############
        e2_enc_outputs = [inputs[k]["encoder_output"] for k in self.entity_2]

        # ================ Flatten ================
        batch_size = e2_enc_outputs[0].shape[0]
        e2_enc_outputs = [torch.reshape(eo, [batch_size, -1]) for eo in e2_enc_outputs]

        # ================ Concat ================
        if len(e2_enc_outputs) > 1:
            e2_hidden = torch.cat(e2_enc_outputs, 1)
        else:
            e2_hidden = list(e2_enc_outputs)[0]

        # ================ Fully Connected ================
        e2_hidden = self.e2_fc_stack(e2_hidden)  # [bs, output_size]

        ###########
        # Compare #
        ###########
        if e1_hidden.shape != e2_hidden.shape:
            raise ValueError(
                f"Mismatching shapes among dimensions! "
                f"entity1 shape: {e1_hidden.shape} "
                f"entity2 shape: {e2_hidden.shape}"
            )

        element_wise_mul = e1_hidden * e2_hidden  # [bs, output_size]
        dot_product = torch.sum(element_wise_mul, 1, keepdim=True)  # [bs, 1]
        abs_diff = torch.abs(e1_hidden - e2_hidden)  # [bs, output_size]
        bilinear_prod = torch.bmm(
            torch.mm(e1_hidden, self.bilinear_weights).unsqueeze(1), e2_hidden.unsqueeze(-1)
        ).squeeze(
            -1
        )  # [bs, 1]

        logger.debug(
            "preparing combiner output by concatenating these tensors: "
            f"dot_product: {dot_product.shape}, element_size_mul: {element_wise_mul.shape}"
            f", abs_diff: {abs_diff.shape}, bilinear_prod {bilinear_prod.shape}"
        )
        hidden = torch.cat([dot_product, element_wise_mul, abs_diff, bilinear_prod], 1)  # [bs, 2 * output_size + 2]

        return {"combiner_output": hidden}

    @staticmethod
    def get_schema_cls():
        return ComparatorCombinerConfig


@register_combiner(name="project_aggregate")
class ProjectAggregateCombiner(Combiner):
    def __init__(
        self, input_features: Dict[str, "InputFeature"] = None, config: ProjectAggregateCombinerConfig = None, **kwargs
    ):
        super().__init__(input_features)
        self.name = "ProjectAggregateCombiner"
        logger.debug(f" {self.name}")

        logger.debug("  Projectors")
        self.projectors = ModuleList(
            # regardless of rank-2 or rank-3 input, torch.prod() calculates size
            # after flattening the encoder output tensor
            [
                Linear(
                    torch.prod(torch.Tensor([*input_features[inp].output_shape])).type(torch.int32),
                    config.projection_size,
                )
                for inp in input_features
            ]
        )

        self.fc_stack = None

        # todo future: this may be redundant, check
        fc_layers = config.fc_layers
        if fc_layers is None and config.num_fc_layers is not None:
            fc_layers = []
            for i in range(config.num_fc_layers):
                fc_layers.append({"output_size": config.output_size})

        self.fc_layers = fc_layers
        if self.fc_layers is not None:
            logger.debug("  FCStack")
            self.fc_stack = FCStack(
                first_layer_input_size=config.projection_size,
                layers=config.fc_layers,
                num_layers=config.num_fc_layers,
                default_output_size=config.output_size,
                default_use_bias=config.use_bias,
                default_weights_initializer=config.weights_initializer,
                default_bias_initializer=config.bias_initializer,
                default_norm=config.norm,
                default_norm_params=config.norm_params,
                default_activation=config.activation,
                default_dropout=config.dropout,
                residual=config.residual,
            )

        if input_features and len(input_features) == 1 and self.fc_layers is None:
            self.supports_masking = True

    def forward(self, inputs: Dict) -> Dict:  # encoder outputs
        encoder_outputs = [inputs[k]["encoder_output"] for k in inputs]

        # ================ Flatten ================
        batch_size = encoder_outputs[0].shape[0]
        encoder_outputs = [torch.reshape(eo, [batch_size, -1]) for eo in encoder_outputs]

        # ================ Project ================
        projected = [self.projectors[i](eo) for i, eo in enumerate(encoder_outputs)]
        hidden = torch.stack(projected)
        hidden = torch.permute(hidden, (1, 0, 2))  # shape [bs, num_eo, h]

        # ================ Aggregate ================
        hidden = torch.mean(hidden, dim=1)

        # ================ Fully Connected ================
        if self.fc_stack is not None:
            hidden = self.fc_stack(hidden)

        return_data = {"combiner_output": hidden}

        if len(inputs) == 1:
            # Workaround for including additional tensors from output of input encoders for
            # potential use in decoders, e.g. LSTM state for seq2seq.
            # TODO(Justin): Think about how to make this communication work for multi-sequence
            # features. Other combiners.
            for key, value in [d for d in inputs.values()][0].items():
                if key != "encoder_output":
                    return_data[key] = value

        return return_data

    @staticmethod
    def get_schema_cls():
        return ProjectAggregateCombinerConfig



from collections import defaultdict, namedtuple
from typing import Dict, List, Optional

import torch
from torch import Tensor
from torchmetrics.metric import Metric

from ludwig.constants import COMBINED, LOSS, NAME, TYPE
from ludwig.modules.metric_registry import get_metric_names_for_type
from ludwig.types import FeatureConfigDict


def sequence_mask(lengths: Tensor, maxlen: Optional[int] = None, dtype=torch.bool) -> Tensor:
    """Implements tf.sequence_mask in torch.

    From https://discuss.pytorch.org/t/pytorch-equivalent-for-tf-sequence-mask/39036/2.
    """
    if maxlen is None:
        maxlen = lengths.max()
    row_vector = torch.arange(0, maxlen, 1).to(lengths.device)
    matrix = torch.unsqueeze(lengths, dim=-1)
    mask = row_vector < matrix

    return mask.type(dtype)


def dynamic_partition(data: Tensor, partitions: Tensor, num_partitions: int) -> List[Tensor]:
    """Implements tf.dynamic_partition in torch.

    From https://discuss.pytorch.org/t/equivalent-of-tf-dynamic-partition/53735.
    """
    assert data.size() == partitions.size()

    # Flatten data into 1D vectors to do partitioning correctly.
    data = data.view(-1)
    partitions = partitions.view(-1)
    result = []
    for i in range(num_partitions):
        result += [data[(partitions == i).nonzero().squeeze(1)]]
    return result


def masked_correct_predictions(targets: Tensor, preds: Tensor, targets_sequence_lengths: Tensor) -> Tensor:
    """Masks out special symbols, and returns tensor of correct predictions.

    Args:
        targets: 2D tensor [batch_size, sequence_length]
        preds: 2D tensor [batch_size, sequence_length]

    Returns:
        1D tensor of all correct predictions.
    """
    correct_preds = preds == targets

    mask = sequence_mask(lengths=targets_sequence_lengths, maxlen=correct_preds.shape[1], dtype=torch.int32)
    _, masked_correct_preds = dynamic_partition(data=correct_preds, partitions=mask, num_partitions=2)

    return masked_correct_preds.type(torch.float32)


def get_scalar_from_ludwig_metric(metric: Metric) -> float:
    """Returns the scalar value of a Ludwig metric.

    Params:
        metric: Metric object

    Returns:
        float: scalar value of the metric
    """
    return metric.compute().detach().cpu().numpy().item()


# Data for training and evaluation metrics.
TrainerMetric = namedtuple("TrainerMetric", ("epoch", "step", "value"))


def reduce_trainer_metrics_dict(
    dict_dict_trainer_metrics: Dict[str, Dict[str, List[TrainerMetric]]]
) -> Dict[str, Dict[str, List[float]]]:
    """Reduces Dict[feature_name, Dict[metric_name, List[TrainerMetric]]] to Dict[feature_name, Dict[metric_name,
    List[float]]].

    Used for flattening the results returned by trainer.py::train(), which come from ProgressTracker.
    """
    flattened_dict = defaultdict(lambda: defaultdict(list))
    for feature_name, trainer_metric_dict in dict_dict_trainer_metrics.items():
        for metric_name, trainer_metrics in trainer_metric_dict.items():
            for trainer_metric in trainer_metrics:
                flattened_dict[feature_name][metric_name].append(trainer_metric[-1])
    # Convert defaultdict to dict so JSON serialization works with dataclasses.asdict().
    return {k: dict(v) for k, v in flattened_dict.items()}


def get_metric_names(output_features: Dict[str, "OutputFeature"]) -> Dict[str, List[str]]:  # noqa
    """Returns a dict of output_feature_name -> list of metric names."""
    metrics_names = {}
    for output_feature_name, output_feature in output_features.items():
        metrics_names[output_feature_name] = sorted(list(get_metric_names_for_type(output_feature.type())))
    # Add combined loss.
    metrics_names[COMBINED] = [LOSS]
    return metrics_names


def get_feature_to_metric_names_map(output_features: List[FeatureConfigDict]) -> Dict[str, List[str]]:
    """Returns a dict of output_feature_name -> list of metric names."""
    metrics_names = {}
    for output_feature in output_features:
        output_feature_name = output_feature[NAME]
        output_feature_type = output_feature[TYPE]
        metrics_names[output_feature_name] = get_metric_names_for_type(output_feature_type)
    metrics_names[COMBINED] = [LOSS]
    return metrics_names


def get_feature_to_metric_names_map_from_feature_collection(
    output_features: "FeatureCollection",  # noqa
) -> Dict[str, List[str]]:
    """Returns a dict of output_feature_name -> list of metric names."""
    metrics_names = {
        output_feature.name: get_metric_names_for_type(output_feature.type) for output_feature in output_features
    }
    metrics_names[COMBINED] = [LOSS]
    return metrics_names



#! /usr/bin/env python
# Copyright (c) 2020 Uber Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================


def get_pandas_dataset_manager(**kwargs):
    from ludwig.data.dataset.pandas import PandasDatasetManager

    return PandasDatasetManager(**kwargs)


def get_ray_dataset_manager(**kwargs):
    from ludwig.data.dataset.ray import RayDatasetManager

    return RayDatasetManager(**kwargs)


dataset_registry = {
    "hdf5": get_pandas_dataset_manager,
    "ray": get_ray_dataset_manager,
    None: get_pandas_dataset_manager,
}


def create_dataset_manager(backend, cache_format, **kwargs):
    return dataset_registry[cache_format](backend=backend, **kwargs)



from collections import defaultdict

from ludwig.api_annotations import DeveloperAPI
from ludwig.constants import MODEL_ECD, MODEL_GBM
from ludwig.schema import utils as schema_utils
from ludwig.utils.registry import Registry

input_config_registries = defaultdict(Registry)
ecd_input_config_registry = input_config_registries[MODEL_ECD]
gbm_input_config_registry = input_config_registries[MODEL_GBM]

input_mixin_registry = Registry()
output_config_registry = Registry()
output_mixin_registry = Registry()

defaults_config_registry = Registry()


def input_config_registry(model_type: str) -> Registry:
    return input_config_registries[model_type]


@DeveloperAPI
def get_input_feature_cls(name: str):
    # TODO(travis): not needed once we remove existing model config implementation
    return input_config_registries[MODEL_ECD][name]


@DeveloperAPI
def get_output_feature_cls(name: str):
    return output_config_registry[name]


@DeveloperAPI
def get_input_feature_jsonschema(model_type: str):
    """This function returns a JSON schema structured to only requires a `type` key and then conditionally applies
    a corresponding input feature's field constraints.

    Returns: JSON Schema
    """
    input_feature_types = sorted(list(input_config_registry(model_type).keys()))
    schema = {
        "type": "object",
        "properties": {
            "name": {"type": "string", "title": "name", "description": "Name of the input feature."},
            "type": {
                "type": "string",
                "enum": input_feature_types,
                "title": "type",
                "description": "Type of the input feature",
            },
            "column": {"type": "string", "title": "column", "description": "Name of the column."},
        },
        "uniqueItemProperties": ["name"],
        "additionalProperties": True,
        "allOf": get_input_feature_conds(model_type),
        "required": ["name", "type"],
        "title": "input_feature",
    }

    return schema


@DeveloperAPI
def get_input_feature_conds(model_type: str):
    """This function returns a list of if-then JSON clauses for each input feature type along with their properties
    and constraints.

    Returns: List of JSON clauses
    """
    input_feature_types = sorted(list(input_config_registry(model_type).keys()))
    conds = []
    for feature_type in input_feature_types:
        schema_cls = get_input_feature_cls(feature_type)
        feature_schema = schema_utils.unload_jsonschema_from_marshmallow_class(schema_cls)
        feature_props = feature_schema["properties"]
        schema_utils.remove_duplicate_fields(feature_props)

        feature_cond = schema_utils.create_cond({"type": feature_type}, feature_props)
        conds.append(feature_cond)
    return conds


@DeveloperAPI
def get_output_feature_jsonschema(model_type: str):
    """This function returns a JSON schema structured to only requires a `type` key and then conditionally applies
    a corresponding output feature's field constraints.

    Returns: JSON Schema
    """
    output_feature_types = sorted(list(output_config_registry.keys()))
    schema = {
        "type": "object",
        "properties": {
            "name": {"type": "string", "title": "name", "description": "Name of the output feature."},
            "type": {
                "type": "string",
                "enum": output_feature_types,
                "title": "type",
                "description": "Type of the output feature",
            },
            "column": {"type": "string", "title": "column", "description": "Name of the column."},
        },
        "additionalProperties": True,
        "allOf": get_output_feature_conds(),
        "required": ["name", "type"],
        "title": "output_feature",
    }

    return schema


@DeveloperAPI
def get_output_feature_conds():
    """This function returns a list of if-then JSON clauses for each output feature type along with their
    properties and constraints.

    Returns: List of JSON clauses
    """
    output_feature_types = sorted(list(output_config_registry.keys()))
    conds = []
    for feature_type in output_feature_types:
        schema_cls = get_output_feature_cls(feature_type)
        feature_schema = schema_utils.unload_jsonschema_from_marshmallow_class(schema_cls)
        feature_props = feature_schema["properties"]
        schema_utils.remove_duplicate_fields(feature_props)
        feature_cond = schema_utils.create_cond({"type": feature_type}, feature_props)
        conds.append(feature_cond)
    return conds



# Copyright (c) 2021 Linux Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from typing import Tuple, Union

import torch
import torch.nn as nn

from ludwig.utils.torch_utils import LudwigModule


class MLP(LudwigModule):
    def __init__(
        self,
        in_features: Union[int, Tuple[int]],
        hidden_size: int,
        out_features: Union[int, Tuple[int]] = None,
        dropout: float = 0.0,
    ):
        super().__init__()

        out_features = out_features or in_features

        self._input_shape = in_features
        self._output_shape = out_features

        self.linear1 = nn.Linear(in_features=in_features, out_features=hidden_size)
        self.linear2 = nn.Linear(in_features=hidden_size, out_features=out_features)
        self.dropout1 = nn.Dropout(dropout)
        self.dropout2 = nn.Dropout(dropout)

    def forward(self, inputs, **kwargs):
        hidden = self.dropout1(nn.functional.gelu(self.linear1(inputs)))
        return self.dropout2(self.linear2(hidden))

    @property
    def input_shape(self) -> torch.Size:
        return torch.Size([self._input_shape])

    @property
    def output_shape(self) -> torch.Size:
        return torch.Size([self._output_shape])


class MixerBlock(LudwigModule):
    def __init__(self, embed_size: int, n_patches: int, token_dim: int, channel_dim: int, dropout: float = 0.0):
        super().__init__()
        self._input_shape = (n_patches, embed_size)
        self._output_shape = (n_patches, embed_size)

        self.mlp1 = MLP(in_features=n_patches, hidden_size=token_dim, dropout=dropout)

        self.mlp2 = MLP(in_features=embed_size, hidden_size=channel_dim, dropout=dropout)

        self.layernorm1 = nn.LayerNorm(normalized_shape=embed_size)
        self.layernorm2 = nn.LayerNorm(normalized_shape=embed_size)

    def forward(self, inputs: torch.Tensor, **kwargs):
        assert inputs.shape[1:] == self.input_shape

        hidden = inputs
        hidden = self.layernorm1(hidden).transpose(1, 2)
        hidden = self.mlp1(hidden).transpose(1, 2)

        mid = hidden + inputs

        hidden = self.layernorm2(mid)
        hidden = self.mlp2(hidden)

        output = hidden + mid
        assert output.shape[1:] == self.output_shape
        return output

    @property
    def input_shape(self) -> torch.Size:
        return torch.Size(self._input_shape)

    @property
    def output_shape(self) -> torch.Size:
        return torch.Size(self._output_shape)


class MLPMixer(LudwigModule):
    """MLPMixer.

    Implements
    MLP-Mixer: An all-MLP Architecture for Vision
    https://arxiv.org/abs/2105.01601
    """

    def __init__(
        self,
        img_height: int,
        img_width: int,
        in_channels: int,
        patch_size: int = 16,
        embed_size: int = 512,
        token_size: int = 2048,
        channel_dim: int = 256,
        num_layers: int = 8,
        dropout: float = 0.0,
        avg_pool: bool = True,
    ):
        super().__init__()
        assert (img_height % patch_size == 0) and (img_width % patch_size == 0)

        self._input_shape = (in_channels, img_height, img_width)
        n_patches = int(img_height * img_width / (patch_size**2))

        self.patch_conv = nn.Conv2d(
            in_channels=in_channels, out_channels=embed_size, kernel_size=patch_size, stride=patch_size
        )

        self.mixer_blocks = nn.ModuleList(
            [
                MixerBlock(
                    embed_size=embed_size,
                    n_patches=n_patches,
                    token_dim=token_size,
                    channel_dim=channel_dim,
                    dropout=dropout,
                )
                for _ in range(num_layers)
            ]
        )

        self.layer_norm = nn.LayerNorm(normalized_shape=embed_size)

        self.avg_pool = avg_pool
        if self.avg_pool:
            self._output_shape = torch.Size((embed_size,))
        else:
            self._output_shape = torch.Size((n_patches, embed_size))

    def forward(self, inputs: torch.Tensor) -> torch.Tensor:
        assert inputs.shape[1:] == self.input_shape
        hidden = self.patch_conv(inputs)
        hidden = hidden.flatten(2).transpose(1, 2)

        for mixer_block in self.mixer_blocks:
            hidden = mixer_block(hidden)
        hidden = self.layer_norm(hidden)

        if self.avg_pool:
            hidden = torch.mean(hidden, dim=1)

        assert hidden.shape[1:] == self.output_shape

        return hidden

    @property
    def input_shape(self) -> torch.Size:
        return torch.Size(self._input_shape)

    @property
    def output_shape(self) -> torch.Size:
        return self._output_shape



import logging
from dataclasses import Field, field
from typing import Any, Dict, Generic, Iterable, List, Optional, Tuple, TypeVar

from marshmallow import fields, validate
from rich.console import Console

from ludwig.api_annotations import DeveloperAPI
from ludwig.constants import (
    AUDIO,
    BAG,
    BINARY,
    CATEGORY,
    DATE,
    H3,
    IMAGE,
    MODEL_ECD,
    MODEL_GBM,
    NUMBER,
    SEQUENCE,
    SET,
    TEXT,
    TIMESERIES,
    VECTOR,
)
from ludwig.schema import utils as schema_utils
from ludwig.schema.features.utils import (
    ecd_input_config_registry,
    gbm_input_config_registry,
    get_input_feature_jsonschema,
    get_output_feature_jsonschema,
    output_config_registry,
)
from ludwig.schema.metadata.parameter_metadata import INTERNAL_ONLY, ParameterMetadata
from ludwig.schema.utils import ludwig_dataclass

logger = logging.getLogger(__name__)
_error_console = Console(stderr=True, style="bold red")
_info_console = Console(stderr=True, style="bold green")


@DeveloperAPI
@ludwig_dataclass
class BaseFeatureConfig(schema_utils.BaseMarshmallowConfig):
    """Base class for feature configs."""

    active: bool = True

    name: str = schema_utils.String(
        default=None,
        allow_none=True,
        description="Name of the feature.",
    )

    type: str = schema_utils.StringOptions(
        default=None,
        allow_none=True,
        options=[AUDIO, BAG, BINARY, CATEGORY, DATE, H3, IMAGE, NUMBER, SEQUENCE, SET, TEXT, TIMESERIES, VECTOR],
        description="Type of the feature.",
    )

    column: str = schema_utils.String(
        allow_none=True,
        default=None,
        description="The column name of this feature. Defaults to name if not specified.",
    )

    proc_column: str = schema_utils.String(
        allow_none=True,
        default=None,
        description="The name of the preprocessed column name of this feature. Internal only.",
        parameter_metadata=ParameterMetadata(internal_only=True),
    )

    def enable(self):
        """This function allows the user to specify which features from a dataset should be included during model
        training. This is the equivalent to toggling on a feature in the model creation UI.

        Returns:
            None
        """
        if self.active:
            _error_console.print("This feature is already enabled!")
        else:
            self.active = True
            _info_console.print(f"{self.name} feature enabled!\n")
            logger.info(self.__repr__())

    def disable(self):
        """This function allows the user to specify which features from a dataset should not be included during
        model training. This is the equivalent to toggling off a feature in the model creation UI.

        Returns:
            None
        """
        if not self.active:
            _error_console.print("This feature is already disabled!")
        else:
            self.active = False
            _info_console.print(f"{self.name} feature disabled!\n")
            logger.info(self.__repr__())


@DeveloperAPI
@ludwig_dataclass
class BaseInputFeatureConfig(BaseFeatureConfig):
    """Base input feature config class."""

    tied: str = schema_utils.String(
        default=None,
        allow_none=True,
        description="Name of input feature to tie the weights of the encoder with.  It needs to be the name of a "
        "feature of the same type and with the same encoder parameters. If text or sequence features are tied, "
        "consider setting the `sequence_length` parameter in `preprocessing` to ensure that the tied features have "
        "equal sized outputs. This is necessary when using the `sequence` combiner.",
    )

    def has_augmentation(self) -> bool:
        return False


@DeveloperAPI
@ludwig_dataclass
class ECDInputFeatureConfig(BaseFeatureConfig):
    pass


@DeveloperAPI
@ludwig_dataclass
class GBMInputFeatureConfig(BaseFeatureConfig):
    pass


@DeveloperAPI
@ludwig_dataclass
class BaseOutputFeatureConfig(BaseFeatureConfig):
    """Base output feature config class."""

    reduce_input: str = schema_utils.ReductionOptions(
        default="sum",
        description="How to reduce an input that is not a vector, but a matrix or a higher order tensor, on the first "
        "dimension (second if you count the batch dimension)",
    )

    default_validation_metric: str = schema_utils.String(
        default=None,
        allow_none=True,
        description="Internal only use parameter: default validation metric for output feature.",
        parameter_metadata=INTERNAL_ONLY,
    )

    dependencies: List[str] = schema_utils.List(
        default=[],
        description="List of input features that this feature depends on.",
    )

    reduce_dependencies: str = schema_utils.ReductionOptions(
        default="sum",
        description="How to reduce the dependencies of the output feature.",
    )

    input_size: int = schema_utils.PositiveInteger(
        default=None,
        allow_none=True,
        description="Size of the input to the decoder.",
        parameter_metadata=ParameterMetadata(internal_only=True),
    )

    num_classes: int = schema_utils.PositiveInteger(
        default=None,
        allow_none=True,
        description="Size of the input to the decoder.",
        parameter_metadata=ParameterMetadata(internal_only=True),
    )


T = TypeVar("T", bound=BaseFeatureConfig)


class FeatureCollection(Generic[T], schema_utils.ListSerializable):
    def __init__(self, features: List[T]):
        self._features = features
        self._name_to_feature = {f.name: f for f in features}
        for k, v in self._name_to_feature.items():
            setattr(self, k, v)

    def to_list(self) -> List[Dict[str, Any]]:
        out_list = []
        for feature in self._features:
            out_list.append(feature.to_dict())
        return out_list

    def items(self) -> Iterable[Tuple[str, T]]:
        return self._name_to_feature.items()

    def __iter__(self):
        return iter(self._features)

    def __len__(self):
        return len(self._features)

    def __getitem__(self, i) -> T:
        if isinstance(i, str):
            return self._name_to_features[i]
        else:
            return self._features[i]


class FeatureList(fields.List):
    def _serialize(self, value, attr, obj, **kwargs) -> Optional[List[Any]]:
        if value is None:
            return None

        value_list = value.to_list()
        return super()._serialize(value_list, attr, obj, **kwargs)

    def _deserialize(self, value, attr, data, **kwargs) -> FeatureCollection:
        feature_list = super()._deserialize(value, attr, data, **kwargs)
        return FeatureCollection(feature_list)


class FeaturesTypeSelection(schema_utils.TypeSelection):
    def __init__(self, *args, min_length: Optional[int] = 1, max_length: Optional[int] = None, **kwargs):
        super().__init__(*args, **kwargs)
        self.min_length = min_length
        self.max_length = max_length

    def get_list_field(self) -> Field:
        min_length = self.min_length
        max_length = self.max_length
        equal = None
        if min_length == max_length:
            min_length = None
            max_length = None
            equal = self.max_length

        return field(
            metadata={
                "marshmallow_field": FeatureList(
                    self,
                    validate=validate.Length(
                        min=min_length,
                        max=max_length,
                        equal=equal,
                    ),
                )
            },
        )


class ECDInputFeatureSelection(FeaturesTypeSelection):
    def __init__(self):
        super().__init__(registry=ecd_input_config_registry, description="Type of the input feature")

    @staticmethod
    def _jsonschema_type_mapping():
        return get_input_feature_jsonschema(MODEL_ECD)


class GBMInputFeatureSelection(FeaturesTypeSelection):
    def __init__(self):
        super().__init__(registry=gbm_input_config_registry, description="Type of the input feature")

    @staticmethod
    def _jsonschema_type_mapping():
        return get_input_feature_jsonschema(MODEL_GBM)


class ECDOutputFeatureSelection(FeaturesTypeSelection):
    def __init__(self):
        super().__init__(registry=output_config_registry, description="Type of the output feature")

    @staticmethod
    def _jsonschema_type_mapping():
        return get_output_feature_jsonschema(MODEL_ECD)


class GBMOutputFeatureSelection(FeaturesTypeSelection):
    def __init__(self):
        super().__init__(max_length=1, registry=output_config_registry, description="Type of the output feature")

    @staticmethod
    def _jsonschema_type_mapping():
        return get_output_feature_jsonschema(MODEL_GBM)



import logging
import os
import platform
import tempfile
from urllib.parse import quote

import pytest

from ludwig.utils.fs_utils import get_fs_and_path, safe_move_directory

logger = logging.getLogger(__name__)


def create_file(url):
    _, path = get_fs_and_path(url)
    logger.info(f"saving url '{url}' to path '{path}'")
    with tempfile.TemporaryDirectory() as tmpdir:
        file_path = os.path.join(tmpdir, path)
        os.makedirs(os.path.dirname(file_path))
        with open(file_path, "w"):
            return path


@pytest.mark.filesystem
def test_get_fs_and_path_simple():
    assert create_file("http://a/b.jpg") == os.path.join("a", "b.jpg")


@pytest.mark.filesystem
def test_get_fs_and_path_query_string():
    assert create_file("http://a/b.jpg?c=d") == os.path.join("a", "b.jpg")


@pytest.mark.filesystem
def test_get_fs_and_path_decode():
    assert create_file("http://a//b%20c.jpg") == os.path.join("a", "b c.jpg")


@pytest.mark.filesystem
def test_get_fs_and_path_unicode():
    assert create_file("http://a/æ.jpg") == "a/æ.jpg"


@pytest.mark.filesystem
@pytest.mark.skipif(platform.system() == "Windows", reason="Skipping if windows.")
def test_get_fs_and_path_invalid_linux():
    invalid_chars = {
        "\x00": ValueError,
        "/": FileExistsError,
    }
    for c, e in invalid_chars.items():
        url = f"http://a/{quote(c)}"
        with pytest.raises(e):
            create_file(url)


@pytest.mark.filesystem
@pytest.mark.skipif(platform.system() != "Windows", reason="Skipping if not windows.")
def test_get_fs_and_path_invalid_windows():
    invalid_chars = {
        "\x00": ValueError,
        "\\": FileExistsError,
        "/": OSError,
        ":": OSError,
        "*": OSError,
        "?": OSError,
        '"': OSError,
        "<": OSError,
        ">": OSError,
        "|": OSError,
    }
    for c, e in invalid_chars.items():
        url = f"http://a/{quote(c)}"
        with pytest.raises(e):
            create_file(url)


@pytest.mark.filesystem
def test_safe_move_directory(tmpdir):
    src_dir = os.path.join(tmpdir, "src")
    dst_dir = os.path.join(tmpdir, "dst")

    os.mkdir(src_dir)
    os.mkdir(dst_dir)

    with open(os.path.join(src_dir, "file.txt"), "w") as f:
        f.write("test")

    safe_move_directory(src_dir, dst_dir)

    assert not os.path.exists(src_dir)
    assert os.path.exists(os.path.join(dst_dir, "file.txt"))
    with open(os.path.join(dst_dir, "file.txt")) as f:
        assert f.read() == "test"



# Copyright (c) 2022 Predibase, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import numpy as np
import pandas as pd

from ludwig.constants import SPLIT
from ludwig.datasets.loaders.dataset_loader import DatasetLoader


class RandomSplitLoader(DatasetLoader):
    """Adds a random split column to the dataset, with fixed proportions of:
     train: 70%
     validation: 10%
     test: 20%
    ."""

    def transform_dataframe(self, dataframe: pd.DataFrame) -> pd.DataFrame:
        df = super().transform_dataframe(dataframe)
        df[SPLIT] = np.random.choice(3, len(df), p=(0.7, 0.1, 0.2)).astype(np.int8)
        return df



import torch


def rmspe_loss(targets: torch.Tensor, predictions: torch.Tensor) -> torch.Tensor:
    """Root mean square percentage error.

    Bad predictions can lead to arbitrarily large RMSPE values, especially if some values of targets are very close to
    zero. We return a large value instead of inf when (some) targets are zero.
    """
    epsilon = 1e-4
    # add epsilon if targets are zero to avoid division by zero
    denominator = targets + epsilon * (targets == 0).float()
    loss = torch.sqrt(torch.mean(((targets - predictions).float() / denominator) ** 2))
    return loss


def mean_confidence_penalty(probabilities: torch.Tensor, num_classes: int) -> torch.Tensor:
    max_entropy = torch.log(torch.tensor(num_classes))
    # clipping needed for avoiding log(0) = -inf
    entropy_per_class, _ = torch.max(-probabilities * torch.log(torch.clamp(probabilities, 1e-10, 1)), dim=0)
    entropy = torch.sum(entropy_per_class, -1)
    penalty = (max_entropy - entropy) / max_entropy
    return torch.mean(penalty)



import numpy as np
import pytest

from ludwig.utils import calibration


@pytest.fixture
def uncalibrated_logits_and_labels():
    """Returns a pair of logits (10x3) and labels (10)."""
    return (
        np.array(
            [
                [-3.596756, 6.728981, 6.3807454],
                [-16.818138, -3.5217745, -1.7786252],
                [-16.060827, 4.7207646, 3.5336719],
                [-4.784969, 5.062503, 3.515455],
                [-4.669478, 7.171067, 6.5137157],
                [-32.596764, -3.5582566, -5.2003713],
                [-4.4035864, 6.3911495, 4.7273974],
                [-4.2035627, 7.846533, 6.0476217],
                [-20.748848, -3.1521742, -4.873552],
                [-4.8901286, 4.726167, 3.208372],
            ]
        ),
        np.array([2, 0, 2, 1, 1, 2, 0, 1, 0, 1]),
    )


EPSILON = 0.1  # maximum relative precision error allowed.


def test_temperature_scaling_binary(uncalibrated_logits_and_labels):
    logits, labels = uncalibrated_logits_and_labels
    # Selects one category of the 3-class test fixture to treat as a binary classifier.
    binary_logits = logits[:, 1]
    binary_labels = labels == 1
    temperature_scaling = calibration.TemperatureScaling(binary=True)
    calibration_result = temperature_scaling.train_calibration(binary_logits, binary_labels)
    # Checks that we got close to optimal temperature
    assert temperature_scaling.temperature.item() == pytest.approx(8.3, EPSILON)
    # Checks that negative log-likelhood and expected calibration error are the same or lower post-calibration.
    assert calibration_result.after_calibration_nll <= calibration_result.before_calibration_nll
    assert calibration_result.after_calibration_ece <= calibration_result.before_calibration_ece


def test_temperature_scaling_category(uncalibrated_logits_and_labels):
    logits, labels = uncalibrated_logits_and_labels
    temperature_scaling = calibration.TemperatureScaling(num_classes=logits.shape[-1])
    calibration_result = temperature_scaling.train_calibration(logits, labels)
    # Checks that we got close to optimal temperature
    assert temperature_scaling.temperature.item() == pytest.approx(19.1, EPSILON)
    # Checks that negative log-likelhood and expected calibration error are the same or lower post-calibration.
    assert calibration_result.after_calibration_nll <= calibration_result.before_calibration_nll
    assert calibration_result.after_calibration_ece <= calibration_result.before_calibration_ece


def test_matrix_scaling_category(uncalibrated_logits_and_labels):
    logits, labels = uncalibrated_logits_and_labels
    matrix_scaling = calibration.MatrixScaling(num_classes=logits.shape[-1])
    calibration_result = matrix_scaling.train_calibration(logits, labels)
    # Matrix scaling may not have a single optimum, so multiple runs could give different results.
    # In this case we don't check any specific values
    # Checks that negative log-likelhood and expected calibration error are the same or lower post-calibration.
    assert calibration_result.after_calibration_nll <= calibration_result.before_calibration_nll
    assert calibration_result.after_calibration_ece <= calibration_result.before_calibration_ece



# Copyright (c) 2022 Predibase, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import calendar
import os
from typing import List

import numpy as np
import pandas as pd

from ludwig.datasets.loaders.dataset_loader import DatasetLoader


class RossmanStoreSalesLoader(DatasetLoader):
    """The Rossmann Store Sales dataset."""

    def load_unprocessed_dataframe(self, file_paths: List[str]) -> pd.DataFrame:
        """Load dataset files into a dataframe."""

        stores_df = pd.read_csv(os.path.join(self.raw_dataset_dir, "store.csv"))

        train_df = pd.read_csv(os.path.join(self.raw_dataset_dir, "train.csv"), low_memory=False)
        train_df = preprocess_df(train_df, stores_df)

        train_df["split"] = -1
        train_df.loc[train_df["Year"] == 2014, "split"] = 0
        train_df.loc[train_df["Year"] == 2015, "split"] = 2
        train_df.drop(train_df[train_df["split"] == -1].index, inplace=True)
        return train_df


def preprocess_dates(df):
    # Make integer Year,Month,Day columns instead of Date
    dates = np.array([[int(v) for v in s.split("-")] for s in df["Date"]])
    df = df.drop(["Date"], axis=1)
    df["Year"] = dates[:, 0]
    df["Month"] = dates[:, 1]
    df["Day"] = dates[:, 2]
    return df


month_abbrs = calendar.month_abbr[1:]
month_abbrs[8] = "Sept"


def preprocess_stores(df, stores_df):
    # join data in df with stores df
    df = df.join(stores_df, on="Store", rsuffix="_right")
    df = df.drop(["Store_right"], axis=1)

    promo2_start_months = [(s.split(",") if not pd.isnull(s) else []) for s in df["PromoInterval"]]

    for month_abbr in month_abbrs:
        df["Promo2Start_" + month_abbr] = np.array(
            [(1 if month_abbr in s else 0) for s in promo2_start_months], dtype=np.int8
        )
    df = df.drop(["PromoInterval"], axis=1)

    return df


int_columns = [
    "Store",
    "DayOfWeek",
    "Sales",
    "Customers",
    "Open",
    "Promo",
    "SchoolHoliday",
    "Year",
    "Month",
    "Day",
    "CompetitionDistance",
    "CompetitionOpenSinceMonth",
    "CompetitionOpenSinceYear",
    "Promo2",
    "Promo2SinceWeek",
    "Promo2SinceYear",
]


def preprocess_df(df, stores_df):
    df = preprocess_dates(df)
    df = preprocess_stores(df, stores_df)

    for column in int_columns:
        df[column] = pd.to_numeric(df[column].fillna(0), downcast="integer")

    df["StateHoliday"] = df["StateHoliday"].astype(str)
    df.loc[df["StateHoliday"] == "0", "StateHoliday"] = "No"

    return df



import json
import os
from typing import Optional, Type, Union

import pytest
import torch

import ludwig.schema.encoders.utils as schema_encoders_utils
from ludwig.api import LudwigModel
from ludwig.constants import ENCODER, MODEL_ECD, NAME, TEXT, TRAINER
from ludwig.encoders import text_encoders
from ludwig.globals import MODEL_HYPERPARAMETERS_FILE_NAME
from ludwig.schema.model_config import ModelConfig
from ludwig.utils.data_utils import load_json
from tests.integration_tests.parameter_update_utils import check_module_parameters_updated
from tests.integration_tests.utils import category_feature, generate_data, HF_ENCODERS, LocalTestBackend, text_feature


def _load_pretrained_hf_model_no_weights(
    modelClass: Type,
    pretrained_model_name_or_path: Optional[Union[str, os.PathLike]],
    **pretrained_kwargs,
):
    """Loads a HF model architecture without loading the weights."""
    from transformers import AutoConfig, AutoModel

    config = AutoConfig.from_pretrained(pretrained_model_name_or_path)
    return AutoModel.from_config(config)


@pytest.fixture
def mock_load_encoder_from_hf_hub(monkeypatch):
    """Mocks encoder downloads from HuggingFace Hub.

    With this mock, only encoder configs are downloaded, not the encoder weights.
    """
    monkeypatch.setattr(text_encoders, "load_pretrained_hf_model", _load_pretrained_hf_model_no_weights)


def get_mismatched_config_params(ludwig_results_dir, ludwig_model):
    saved_config_dict = load_json(os.path.join(ludwig_results_dir, "model", MODEL_HYPERPARAMETERS_FILE_NAME))
    saved_config_obj = ModelConfig.from_dict(saved_config_dict)

    mismatches = []
    for input_feature_config in saved_config_obj.input_features.to_list():
        feature_name = input_feature_config[NAME]
        encoder_config_from_file = input_feature_config[ENCODER]
        encoder_config_from_model = ludwig_model.model.input_features.get(feature_name).encoder_obj.config.to_dict()
        for k, v in encoder_config_from_model.items():
            # Skip saved_weights_in_checkpoint because this value is not yet set when the global config
            # is modified with the final encoder config.
            if k == "saved_weights_in_checkpoint":
                continue

            if encoder_config_from_file[k] != v:
                mismatch = {
                    "feature_name": feature_name,
                    "param_name": k,
                    "val_from_file": encoder_config_from_file[k],
                    "val_from_model": v,
                }
                mismatches.append(mismatch)
    return mismatches


@pytest.mark.slow
@pytest.mark.parametrize("encoder_name", HF_ENCODERS)
def test_hf_ludwig_model_e2e(tmpdir, csv_filename, mock_load_encoder_from_hf_hub, encoder_name):
    """Tests HuggingFace encoders end-to-end.

    This test validates the following:
        1. Encoder config defaults are compatible with Ludwig experiments.
        2. Ludwig correctly updates the encoder config with the parameters introduced by the HF encoder.
        3. Ludwig correctly loads checkpoints containing HF encoder weights.
    """
    input_features = [
        text_feature(
            encoder={
                "vocab_size": 30,
                "min_len": 1,
                "type": encoder_name,
                "use_pretrained": True,
            }
        )
    ]
    output_features = [category_feature(decoder={"vocab_size": 2})]
    rel_path = generate_data(input_features, output_features, csv_filename)

    config = {
        "input_features": input_features,
        "output_features": output_features,
        TRAINER: {"train_steps": 1},
    }
    model = LudwigModel(config=config, backend=LocalTestBackend())

    # Validates that the defaults associated with the encoder are compatible with Ludwig training.
    _, _, _, results_dir = model.experiment(dataset=rel_path, output_directory=tmpdir)

    # Validate that the saved config reflects the parameters introduced by the HF encoder.
    # This ensures that the config updates after initializing the encoder.
    mismatched_config_params = get_mismatched_config_params(results_dir, model)
    if len(mismatched_config_params) > 0:
        raise AssertionError(
            f"Config parameters mismatched with encoder parameters: {json.dumps(mismatched_config_params, indent=4)}"
        )

    # Validate the model can be loaded.
    # This ensures that the config reflects the internal architecture of the encoder.
    LudwigModel.load(os.path.join(results_dir, "model"))


@pytest.mark.parametrize("trainable", [True, False])
def test_distilbert_param_updates(trainable: bool):
    max_sequence_length = 20
    distil_bert_encoder = text_encoders.DistilBERTEncoder(
        use_pretrained=False,
        max_sequence_length=max_sequence_length,
        trainable=trainable,
    )

    # send a random input through the model with its initial weights
    inputs = torch.rand((2, max_sequence_length)).type(distil_bert_encoder.input_dtype)
    outputs = distil_bert_encoder(inputs)

    # perform a backward pass to update the model params
    target = torch.randn(outputs["encoder_output"].shape)
    check_module_parameters_updated(distil_bert_encoder, (inputs,), target)

    # send the same input through the model again. should be different if trainable, else the same
    outputs2 = distil_bert_encoder(inputs)

    encoder_output1 = outputs["encoder_output"]
    encoder_output2 = outputs2["encoder_output"]

    if trainable:
        # Outputs should be different if the model was updated
        assert not torch.equal(encoder_output1, encoder_output2)
    else:
        # Outputs should be the same if the model wasn't updated
        assert torch.equal(encoder_output1, encoder_output2)


@pytest.mark.parametrize("encoder_name", HF_ENCODERS)
def test_encoder_names_constant_synced_with_schema(encoder_name):
    """Ensures that each value in the HF_ENCODERS constant is represented by an equivalent schema object."""
    schema_encoders_utils.get_encoder_cls(MODEL_ECD, TEXT, encoder_name)
